from opcua import Client as UaClient
from opcua import ua
from collections import defaultdict 
import weakref 

DEFAULT_NAME_SPACE = 4

#   ___  ____   ____      _   _   _
#  / _ \|  _ \ / ___|    | | | | / \
# | | | | |_) | |   _____| | | |/ _ \
# | |_| |  __/| |__|_____| |_| / ___ \
#  \___/|_|    \____|     \___/_/   \_\
#
DEBUG = False

def _name_attr(cl):
    names = {}
    txt = {}
    for subcl in cl.__mro__[::-1]:
        names.update( {n:k for k,n in subcl.__dict__.items() if not k.startswith('_') and isinstance(n, int)} )
        txt.update(getattr(cl, 'txt', {}))
    #cl.names = {n:k for k,n in cl.__dict__.items() if not k.startswith('_')}
    cl.names = names
    cl.txt = {**names, **txt}
    # for c,n in cl.names.items():
    #     setattr( cl, 'is_'+n, classmethod(lambda cl, v, _c_=c: v==_c_))
    return cl

_ua_clients = {}
_ua_clients_references = {}
def get_ua_client(uri):
    """ get an opc-ua client object for the given address create one if not exists """
    return UaClient(uri)
    # ## This was an attempt to share client with devices with same IP 
    # ## But this seems to be problematic
    # global _ua_clients
    # 
    # try:
    #     client = _ua_clients[uri]
    # except KeyError:
    #     if DEBUG:
    #         client = DebugClient(uri)
    #     else:
    #         client = UaClient(uri)
    #     _ua_clients[uri] = client
    # finally:
    #     _ua_clients_references.setdefault(client, {})
    # return client

def connect_ua_client(ua_client, obj):
    ua_client.connect()
    # return 
    # dc = _ua_clients_references.setdefault(ua_client, {})
    # if id(obj) not in dc:
    #     dc[id(obj)] = weakref.ref(obj)
    # ua_client.connect() # I think we can do it without problem
        
def disconnect_ua_client(ua_client, obj):
    ua_client.disconnect()
    return 
    # if not ua_client in _ua_clients_references:
    #     return
    # 
    # dc = _ua_clients_references.setdefault(ua_client, {})
    # try:
    #     dc.pop(id(obj))
    # except KeyError:
    #     pass
    # # do some clean up, remove objects that 
    # # does not exists anymore. It may be time consuming 
    # # but I am assuming that disconnection is not done in 
    # # real time.
    # for uid, ref in list(dc.items()):
    #     if ref() is None:
    #         dc.pop(uid)
    # 
    # if not len(dc):        
    #     ua_client.disconnect()
    #     _ua_clients_references.pop(ua_client)
    
    
    
####
# Some dummy classes to work without any opc-ua server available
# This is just to debug part of the soft this is not a simulator
class DebugClient:
    log = staticmethod(print)
    def __init__(self, url):
        self.url = url
        self.values = {}
    def get_node(self, s):
        return DebugNode(s, self.values)
    def connect(self):
        self.log('connect {}'.format(self.url))
    def disconnect(self):
        self.log('disconnect {}'.format(self.url))

class DebugNode:
    log = staticmethod(print)
    def __init__(self, nodeid, values):
        self.nodeid = nodeid
        self.values = values
    def get_value(self):
        return self.values.setdefault(self.nodeid,0.0)
    def set_value(self, val):
        self.values[self.nodeid] = val
    def set_attribute(self, attrid, uaData):
        self.values[self.nodeid] = uaData.Value.Value
    def call_method(self, mid, *args):
        print( *((a, type(a)) for a in args) )
        return 0


def Int16(value):
    return ua.Variant(int(value), ua.VariantType.Int16) 
INT = Int16

def Int32(value):
    return ua.Variant(int(value), ua.VariantType.Int32)
DINT = Int32

def Int64(value):
    return ua.Variant(int(value), ua.VariantType.Int64)  
LINT = Int64

def UInt16(value):
    return ua.Variant(int(value), ua.VariantType.UInt16) 
UINT = UInt16

def UInt32(value):
    return ua.Variant(int(value), ua.VariantType.UInt32)
UDINT = UInt32

def UInt64(value):
    return ua.Variant(int(value), ua.VariantType.UInt64)        
ULINT = UInt64

def Float(value):
    return ua.Variant(int(value), ua.VariantType.Float)
REAL = Float

def Double(value):
    return ua.Variant(int(value), ua.VariantType.Double)
LREAL = Double
 




#  ___ _   _ _____ _____ ____  _____ _    ____ _____ ____
# |_ _| \ | |_   _| ____|  _ \|  ___/ \  / ___| ____/ ___|
#  | ||  \| | | | |  _| | |_) | |_ / _ \| |   |  _| \___ \
#  | || |\  | | | | |___|  _ <|  _/ ___ \ |___| |___ ___) |
# |___|_| \_| |_| |_____|_| \_\_|/_/   \_\____|_____|____/
#

def kjoin(*args):
    """ join key elements """
    return ".".join(a for a in args if a)

def fjoin(*args):
    """ join fits elements """
    return " ".join(a.strip() for a in args if a)

def ksplit(key):
    s, _, p = key[::-1].partition(".")
    return p[::-1], s[::-1]

def fsplit(key):
    s, _, p = key[::-1].partition(" ")
    return p[::-1], s[::-1]

def setitem(obj,k,v):
    """ setitem(obj, k, v) -> obj[k] = v"""
    obj[k] = v

class RpcError(RuntimeError):
    """ Raised when an rpc method is returning somethingelse than 0

        See rcall method of RpcNode
    """
    rpc_error = 0

## generic, default RPC_ERROR code

### ##############
# RPC error
# generic, default RPC_ERROR codes

@_name_attr
class RPC_ERROR:
    OK =  0
    NOT_OP =  -1
    NOT_NOTOP_READY =  -2
    NOT_NOTOP_NOTREADY = -3
    LOCAL =  -4
    # etc ...
    txt = {
        OK:					 'OK',
        NOT_OP:				 'Cannot control motor. Not in OP state.',
        NOT_NOTOP_READY:	 'Call failed. Not in NOTOP_READY.',
        NOT_NOTOP_NOTREADY:	 'Call failed. Not in NOTOP_NOTREADY/ERROR.',
        LOCAL:				 'RPC calls not allowed in Local mode.',
        # etc 
    }

def dummy(x):
    return x


class NodeAlias:
    """ NodeAlias mimic a real client Node. 
    
    The NodeAlias object does a little bit of computation to return a value with its `get()` method and 
    thanks to required input nodes.
     
    The NodeAlias cannot be use as such without implementing a `fget` method. This can be done by 
    implementing the fget method on the class or as an argument at init (same for the optionaly the fset method). 
    
    NodeAlias is an abstraction layer, it does not do anything complex but allows to simplify the programming 
    of GUIs for instance.  
    
    But NodeAlias object can be easely created with the @nodealias() decorator
    
    Args:
        key (str): Key of the node
        nodes (list): list of nodes necessary for the alias node. This is required because when the 
                     node alias is used in a Downloader object, the Downloader will automaticaly fetch 
                     those required nodes from server (or orher node alias). 
    Exemple: 
    
    ::
    
        >>> is_inpos_for_test = NodeAlias('is_inpos_for_test', [mgr.motor1.stat.pos_actual], fget=lambda pos: abs(pos-4.56)<0.01)
        >>> is_inpos_for_test.get()
    
    :: 
    
        @nodealias("is_all_standstill", [mgr.motor1.stat.substate, mgr.motor2.stat.substate])
        def is_all_standstill(m1_substate, m2_substate):
            return m1_substate == Motor.SUBSTATE.OP_STANDSTILL and m2_substate == Motor.SUBSTATE.OP_STANDSTILL
    
        >>> is_all_standstill.get()
        True
        
        >>> downloader = Downloader( [is_all_standstill] )
        >>> downloader.download()        
        >>> downloader.data
        {'fcs.motor1.substate': 100,
         'fcs.motor2.substate': 100,
         'is_all_standstill': False}
         
    In the exemple above one can see that the mgr.motor[12].stat.substate has been automatically added 
    to the nodes to be fetched from OPC-UA server(s). 
    
    Here is an exemple of customized NodeAlias
    
    ::
        import numpy as np 
        
        class MinMaxNode(NodeAlias):
            min = +np.inf
            max = -np.inf
            
            def fget(self, pos):
                self.min = min(pos, self.min)
                self.max = max(pos, self.max)
                return ( self.min , self.max )
            
            def reset(self):
                self.min = +np.inf
                self.max = -np.inf
                
        mot1_minmax = MinMaxNode( "minmax",  [mgr.motor1.stat.pos_actual])
                
    .. seealso::  
        :func:`nodealias`
        :func:`nodealiasproperty`
        :class:`NodeAlias`
        
    """
    
    def __init__(self, key, nodes, fget=None, fset=None):
        self._nodes = nodes
        self._key = key
        
        if fget:
            self.fget = fget
        if fset:
            self.fset = fset
        
    @property
    def key(self):
        return self._key
    
    @property
    def prefix(self):
        s,_,p = self.key[::-1].partition(".")
        return p[::-1]
    
    @property
    def name(self):
        s,_,p = self.key[::-1].partition(".")
        return s[::-1]
    
    def get(self, data=None):
        """ get the node alias value from server or from data dictionary if given """
        if data is None:
            values = NodesReader(self._nodes).read()
        else:
            values = [data[n.key] for n in self._nodes]
        return self.fget(*values)
    
    def set(self, value, data=None):
        """ set the node alias value to server or to data dictionary if given """
        values = self.fset(value)
        if data is None:
            NodesWriter(dict(zip(self._nodes, values))).write()
        else:
            for n,v in zip(self._nodes, values):
                data[n] = v
    
    @staticmethod
    def fget(*args):
        raise NotImplementedError("Node alias get function")
    
    @staticmethod
    def fset(*args):
        raise NotImplementedError("Node alias set function")
        
class NodeAliasPropertyChildren(NodeAlias):
    def __init__(self, obj, key, nodes, fget=None, fset=None):
        self._obj = obj
        self._nodes = nodes
        
        if fget:
            self.fget = fget
        if fset:
            self.fset = fset
        self._key = key 
        
    def get(self, data=None):
        if data is None:
            values = NodesReader(self._nodes).read()
        else:
            #values = [data[n.key] for n in self._nodes]
            values = [data[n.key] for n in self._nodes]
        return self.fget(self._obj, *values)
    
    def set(self, value, data=None):
        values = self.fset(self._obj, value)
        if data is None:
            NodesWriter(dict(zip(self._nodes, values))).write()
        else:
            for n,v in zip(self._nodes, values):
                data[n] = v
    
class NodeAliasProperty:
    """ A Node Alias to be used inside a Node Interface object 
    
    This class will be mostly generated by the @nodealiasproperty() decorator. 
    The NodeAliasProperty object must be included in a class having the `get_node` method. 
     `get_node` will be used to fetch the node.
     
    Args:
        name (str): node name the key will be parent_key.name where parent_key is the host object key 
        node_names (lst): list of node names necessary for the nodeAlias. The real nodes will be fetch 
                from parent.get_node(node_name)
    
    Exemple:
    
    copy pasted from device definition:
    
    ::
    
        class DeviceStatInterface(Interface):
            ERROR = ERROR # needed for error_txt alias 
            SUBSTATE = SUBSTATE # needed for substate_txt node alias
            STATE = STATE 
            
            @nodealiasproperty("is_operational", ["state"])
            def is_operational(self, state):
                return state == STATE.OP
            
            @nodealiasproperty("is_ready", ["substate"])
            def is_ready(self, substate):
                return substate == self.SUBSTATE.NOTOP_READY 
     
            @nodealiasproperty("substate_txt", ["substate"])
            def substate_txt(self, substate):
                return self.SUBSTATE.txt.get(substate, "not registered substate")
    
    """
    def __init__(self, name, node_names, fget=None, fset=None):
        self.node_names = node_names
        self.fget = fget
        self.fset = fset
        self.name = name
    
    def __get__(self, obj, cls=None):
        try:
            node_alias = obj.__dict__[self.name]
        except KeyError:
            nodes = [obj.get_node(n) for n in self.node_names]
            node_alias = NodeAliasPropertyChildren(obj, kjoin(obj.key,self.name), nodes, fget=self.fget, fset=self.fset)
            # hopefully self.name shall be the same than the name in class definition 
            # otherwise side effects are expected 
            obj.__dict__[self.name] = node_alias
        else:
            print("BUG: nodealiaspropery of name %s have a different class Attributes")
        return node_alias
    
    def getter(self, func):
        """ decorator for the fget method """
        self.fget = func
        return self
    
    def setter(self, func):
        """ decorator for the fset method """
        self.fset = func
        return self 
    

def nodealias(key, nodes):
    """ A decorator to create a :class:`NodeAlias` object 
    
    Args:
        key (str): Key of the node
        nodes (list): list of nodes necessary for the alias node.
    
    Exemple: 
        
    :: 
    
        @nodealias("is_all_standstill", [mgr.motor1.stat.substate, mgr.motor2.stat.substate])
        def is_all_standstill(m1_substate, m2_substate):
            return m1_substate == Motor.SUBSTATE.OP_STANDSTILL and m2_substate == Motor.SUBSTATE.OP_STANDSTILL
    
        >>> is_all_standstill.get()
        True
        
        
    
    .. seealso::  
    
        :class:`NodeAlias`
    
    """
    alias = NodeAlias(key, nodes)
    def fget_setter(func):
        alias.fget = func
        return alias
    return fget_setter   

def nodealiasproperty(name, node_names):
    """ A :class:`NodeAliasProperty` decorator to be used inside a Node Interface object 
    
    The decorator must be included in a class with the `get_node` method. 
    `get_node` will be used to fetch the node.
     
    Args:
        name (str): node name the key will be parent_key.name where parent_key is the host object key 
        node_names (lst): list of node names necessary for the nodeAlias. The real nodes will be fetch 
                from parent.get_node(node_name)
    
    Exemple:
    
    copy pasted from device definition:
    
    ::
    
        class DeviceStatInterface(Interface):
            ERROR = ERROR # needed for error_txt alias 
            SUBSTATE = SUBSTATE # needed for substate_txt node alias
            STATE = STATE 
            
            @nodealiasproperty("is_operational", ["state"])
            def is_operational(self, state):
                return state == STATE.OP
            
            @nodealiasproperty("is_ready", ["substate"])
            def is_ready(self, substate):
                return substate == self.SUBSTATE.NOTOP_READY 
     
            @nodealiasproperty("substate_txt", ["substate"])
            def substate_txt(self, substate):
                return self.SUBSTATE.txt.get(substate, "not registered substate")
    
    """
    alias_property = NodeAliasProperty(name, node_names)
    return alias_property.getter          


class Node:
    """ Object representing a value node

    This is an interface representing one single value (node) in the OPC-UA server. 
    
    Node will be mostly used transparently inside :class:`Interface` object itself embeded inside 
    a :class:`pydevmgr.Device` object
        e.g. : in  `mgr.motor1.stat.pos_actual`
           
           mgr: is a :class:`pydevmgr.Manager`
           motor1: is a :class:`pydevmgr.Motor` extended from :class:`pydevmgr.Device`
           stat :  is a :class:`pydevmgr.Interface`  (group of nodes with a mapping dictionary)
           pos_actual : is a :class:`pydevmgr.Node`  
    
    Args:
        key (str): The string key representing the node in its context.
                   If key='a.b.c'  'a.b' is the `prefix` and 'c' the `name` of the node
        ua_node (opcua.Node): The opcua client node 
        vtype (callable, optional): A value parser for the set method (e.g. float, int, ...)
                This is handy for nodes representing a e.g. 16bits instegert on the server side. 
                
    Exemple:
    
    ::
      
        import pydevmgr
        import opcua
        
        client = opcua.client('url')
        current = pydevmgr.Node( 'current' client.get_node('ns=4;s=MAIN.lrCurrent') )
        substate =  pydevmgr.Node( 'substate', client.get_node('ns=4;s=MAIN.Motor1.stat.nSubstate'), vtype=pydevmgr.INT )
             
    """
    def __init__(self,  key, ua_node, vtype=None, sid=None):
        self._ua_node = ua_node
        self._key = key
        self._vtype = vtype
        self._sid = sid # server id identifier is used to group nodes together
        self._ua_variant_type = None
    
    def __repr__(self):
        return "<Node  key=%r>"%(self.key)
    
    @property
    def key(self):
        return self._key
    
    @property
    def prefix(self):
        return ksplit(self._key)[0]

    
    @property
    def name(self):
        return ksplit(self._key)[1]
        

    def get(self, data=None):
        """ get the server node value from server 
        
        If data is given the value is taken from this dictionary with its key instead of the server.
        """
        if data is None:
            return self._ua_node.get_value()
        else:
            return data[self.key]    
    
    def _parse_value_for_ua(self, value):
        
        if self._vtype:
            value = self._vtype(value)
        
        
        # this is a security net. In most of situation a python int will not work (LINT are rare in TwinCat FB)
        # this shall normally not be called and should be handled by vtype at device creation
        if isinstance(value, (int,)):
            # remove embiguity between int64 and int32, int16
            # we need to ask the variant_type on the server
            if self._ua_variant_type is None:
                self._ua_variant_type = self._ua_node.get_data_type_as_variant_type()
            datavalue = ua.DataValue(ua.Variant(value, self._ua_variant_type))
        elif isinstance(value, ua.DataValue):
            datavalue = value
        elif isinstance(value, ua.Variant):
            datavalue = ua.DataValue(value)
        else:
            datavalue = ua.DataValue(value)
        return datavalue    
        
    def set(self, value, data=None):
        """ set the server node value to server 
        
        If data is given the value is set to this dictionary with its key instead of the server.
        """
        ###
        #
        if data is None:
            datavalue = self._parse_value_for_ua(value)
            self._ua_node.set_attribute(ua.AttributeIds.Value, datavalue)
        else:
            data[self] = value
    
    
    
class RpcNode:
    """ Object representing an OPC-UA Rpc method 

    It does the interce between a python method and the OPC-UA
    
    Both methods  :func:`RpcNode.call` and :func:`RpcNode.rcall` are calling the server Rpc. The first one is returning 
    an rpc_error code, the second return nothing but raise an :class:`RpcError` exception in case of error.
    
    In rare case user will use the RpcNode as it is automatically created from :class:`RpcInterface` containing 
    all the mapping.
    
    Args:
        key (str): shall be unic in its context  
        ua_node (opcua.Node): an opcua client node (see opcua library)
                   !! This is the node of the structure containing the method
                   
        ua_method_id(str): the ua method id
        atypes (None, lst, optional): a list of callable parser for the method. If given the number 
                of `atypes` shall match the number of arguments in the server Rpc Method. (no chec is done)
    
    Exemple:
        
    ::
        
        import opcua 
        from pydevmgr import RpcNode
        
        client = opcua.Client('..url..')
        move_abs = RpcNode('mot1.move_abs', client.get_node('ns=4;s=MAIN.Motor1'), '4:RPC_MoveAbs', atypes=(float, float))         
    
    """
    RPC_ERROR = RPC_ERROR

    def __init__(self, key, ua_node, ua_method_id, atypes=None):
        self._ua_node = ua_node
        self._ua_method_id = ua_method_id
        self._atypes = atypes
        self._key = key
    
    def __repr__(self):
        return "<RpcNode prefix=%r name=%r>"%(self.prefix, self.name)
    
    @property
    def key(self):
        return self._key
    
    @property
    def prefix(self):
        return ksplit(self._key)[0]

    
    @property
    def name(self):
        return ksplit(self._key)[1]
    
    def get_rpc_error_txt(self, rpc_error):
        """ get a txt description of the rpc_error code 
        
        Args:
            rpc_error (int): rpc error code  
        """
        return self.RPC_ERROR.txt.get(rpc_error,'Unregistered RPC error')
    
    def call(self, *args):
        """ Call the method and return what return the server 
        
        this shall return an integer rpc_error which shall be 0 if success
        """
        return self._ua_node.call_method(self._ua_method_id, *self._parse_args(args))

    def rcall(self, *args):
        """ Call the Rpc Method but raised an exception in case of an error code is returned """
        rpc_error = self.call(*args)
        if rpc_error:
            e = RpcError("RPC ({}): {}".format(rpc_error, self.get_rpc_error_txt(rpc_error)))
            e.rpc_error = rpc_error
            raise e
            
    def _parse_args(self, args):
        if self._atypes:
            return ( tpe(a) for a,tpe in zip(args, self._atypes) )
        return args

    def __call__(self, *args):
        self.rcall(*args)


def getkey(obj):
    return obj.key

class NodesWriter:
    def __init__(self, node_values):
        self._dispatch, self._all_wv = _dispatch_node_to_write(dict(node_values))
    
    def write(self, altdata=None):
        _write_dispatch(self._dispatch, self._all_wv, altdata)
        
def _dispatch_node_to_write(nodes_values):
    disptach = {} # None is for aliases 
    all_wv = {}
    for node, value in nodes_values.items():
        if isinstance(node, NodeAlias):
            node.set(nodes_values)
    
    for node, value in nodes_values.items():
        if not isinstance(node, NodeAlias):
            _add_node_to_dispatch_writer(disptach,  all_wv, node, value)
    return disptach, all_wv

def _add_node_to_dispatch_writer(disptach, all_wv, node, value):
    
    try:
        params = disptach[node._sid]
    except KeyError:
        params =  ua.WriteParameters()
        params.uaclient = node._ua_node.server
        params.NodesToWrite = []
        disptach[node._sid] = params
        
    wv = ua.WriteValue()
    wv.NodeId = node._ua_node.nodeid
    wv.AttributeId = ua.AttributeIds.Value
    wv.vparser= node._parse_value_for_ua
    wv.Value = wv.vparser(value)
    params.NodesToWrite.append(wv)
    all_wv[node.key] = wv

def _write_dispatch(dispatch, all_wv, altdata):
    if altdata:
        for k,v in altdata.items():
            wv = all_wv[k]
            wv.Value = wv.vparser(v)
    
    for params in  dispatch.values():
        result = params.uaclient.write(params)
        for r in result: r.check()

class NodesReader:
    def __init__(self, nodes):
        self._input_nodes = nodes # need to save to remenber the order
        self._dispatch, self._aliases =  _dispatch_node_reader(nodes)
    
    def _read_to(self, data, setfunc):
        _read_dispatch_to_data(data, self._dispatch, self._aliases, setfunc)    
    
    def read(self, data=None, setfunc=setitem):
        if data is None:
            data = {}
            self._read_to(data, setfunc)
            return [data[n.key] for n in self._input_nodes]
        
        self._read_to(data, setfunc)
        return None

            
def _dispatch_node_reader(nodes):
    disptach = {} # None is for aliases 
    aliases = []
    for node in nodes:
        _add_node_to_dispatch_reader(disptach, aliases, node) 
    return disptach, aliases

def _add_node_to_dispatch_reader(disptach, aliases, node):
    
    if isinstance(node, NodeAlias):
        aliases.append( node )
        for n in node._nodes:            
            _add_node_to_dispatch_reader(disptach,  aliases, n)
        return 
    
    try:
        params = disptach[node._sid]
    except KeyError:
        params =  ua.ReadParameters()
        params.uaclient = node._ua_node.server
        params.keys = list()
        disptach[node._sid] = params
        
    rv = ua.ReadValueId()
    rv.NodeId = node._ua_node.nodeid
    rv.AttributeId = ua.AttributeIds.Value
    params.NodesToRead.append(rv)
    params.keys.append(node.key)


def _read_dispatch_to_data(data, dispatch, aliases, setfunc):
    for sid, params in dispatch.items():
        
        result = params.uaclient.read(params)
        for key,r in zip(params.keys, result):
            r.StatusCode.check()
            setfunc(data, key, r.Value.Value)        
    
    # aliases are treated at the end, data should have all necessary real nodes for 
    # the alias 
    # We need to start from the last as Aliases at with lower index can depend 
    # of aliases with higher index
    flags = [False]*len(aliases)
    for i, alias in reversed(list(enumerate(aliases))):
        if not flags[i]:        
            setfunc(data, alias.key, alias.get(data))
            flags[i]= True
  


def getitem(d,k):
    return d[k]


def get_values(uaclient, nodes):
    """ get values of several nodes 
    
    All nodes shall be on the same server
    
    Args:
        uaclient : opcua.Client 
        nodes : list of :class:`Node`
    """
    params = ua.ReadParameters()
    for node in nodes:
        _add_node_to_rparams(params, node)
    result = uaclient.read(params)
    #result[0].StatusCode.check()
    for r in result:
        r.StatusCode.check()
    return [r.Value.Value for r in result]

def _add_node_to_rparams(params, node):
    if isinstance(node, NodeAlias):
        for n in node.nodes:
            _add_node_to_rparams(params, n)
        return 
    rv = ua.ReadValueId()
    rv.NodeId = node._ua_node.nodeid
    rv.AttributeId = ua.AttributeIds.Value
    params.NodesToRead.append(rv)



def set_values(uaclient, node_values):
    """ set values of several nodes 
    
    All nodes shall be on the same server
    
    Args:
        uaclient : opcua.Client 
        node_values : dictionary of :class:`Node`/value pairs 
    """
    # for node, value in node_values:
    #     node.set(value)
    # return 
    for node, value in node_values.items():
        if isinstance(node, NodeAlias):
            node.set(node_values)
    
    params = ua.WriteParameters()
    for node, value in node_values.items():
        if not isinstance(node, NodeAlias):
            _add_node_to_wparams(params, node, value)
    
    result = uaclient.write(params)
    for r in result: r.check()
    
def _add_node_to_wparams(params, node, value):
    attr = ua.WriteValue()
    attr.NodeId = node._ua_node.nodeid
    attr.AttributeId = ua.AttributeIds.Value
    
    datavalue = node._parse_value_for_ua(value)
    attr.Value = datavalue           
     
    params.NodesToWrite.append(attr)


class Interface:
    """ Interface between opc-ua nodes and pydevmgr set of :class:`Nodes` 
    
    Args:
        
        key (str): a unique key defining the interface. This is generally the same key as the host 
                  :class:`Device`. e.g. mgr.motor1.key ==  mgr.motor1.stat.key == mgr.motor1.cfg.key
        ua_client (opcua.Client): opcua client connected to some servers
        map (dict): a dictionnary of name/node_suffix pairs 
            Like:     
            
                ::
                  
                    {'velocity' :     'cfg.lrDefaultVelocity',
                     'max_pos' :      'cfg.lrMaxPosition',
                     'min_pos' :      'cfg.lrMinPosition',
                     'check_inpos' :  'cfg.bCheckInPos', 
                     # etc...
                     }
        ua_prefix (str): The prefix for the OPC-UA path like for instance 'MAIN.Motor1.stat'
        ua_namespace (int,str, optional): The ua name space default is 4
    """
    _ua_client = None
    _map = None
    _ua_prefix = None
    _prefix = None
    _name = None
    
    
    Node = Node

    def __init__(self, key, ua_client, map, ua_prefix, ua_namespace=DEFAULT_NAME_SPACE):
        self._ua_client = ua_client
        self._map = map

        self._ua_prefix = ua_prefix
        self._key = key
        
        self._ua_namespace = ua_namespace


    def __getattr__(self, attr):
        try:
            return self.__dict__[attr]
        except KeyError:
            return self.get_node(attr)

    def __dir__(self):
        return ['get', 'set', 'gat_value_node'] + list(self._map.keys())
        
    @classmethod
    def _get_vtype(cl, key):
        """ look in class __annotations__ to find any parser for a given key """

        for subcl in cl.__mro__:
            try:
                return subcl.__annotations__[key]
            except (KeyError,AttributeError):
                continue
        return None
        
    @property
    def all_nodes(self):
        """ a list nodes of the interface """
        return [self.get_node(k) for k in self.map.keys()]
    
    @property
    def key(self):
        return self._key
    
    @property
    def prefix(self):
        return ksplit(self._key)[0]

    
    @property
    def name(self):
        return ksplit(self._key)[1]
    
    @property
    def map(self):
        return self._map
        
    def _get_ua_id(self, name):
        """ return a node id of a given name in the python space """
        try:
            ua_name = self._map[name]
        except KeyError:
            raise KeyError('Unknown node %r'%name)
        return "ns={};s={}".format(self._ua_namespace, kjoin(self._ua_prefix, ua_name))
        
    def _get_ua_node(self, name):
        uaNode = self._ua_client.get_node(self._get_ua_id(name))
        return uaNode
    
    def clear(self):
        """ clear all cashed intermediate objects """
        for k,v in list(self.__dict__.items()):
            if isinstance(v, (Node, NodeAlias)):
                self.__dict__.pop(k)
    
    def get_nodes(self, node_names=None):
        """ return a list of :class:`Node` from a list of names 
        
        if node_names is None all the native nodes are return. The :class:`NodeAliasProperty` 
        are not included.
        
        
        !!! Warning if a given name startwith '_' this method may return something which is not 
           a :class:`Node`. No check is done for performance raison.  
        """
        node_names = self.map.keys() if node_names is None else node_names
        return [self.get_node(n) for n in node_names]
    
    def get_node(self, name):
        """ Return a Node for the given key

        key must be on of the keys in the mapping file.
        
        !!! Warning if the given name startwith '_' this method may return something which is not 
           a :class:`Node`. No check is done for performance raison.  

        Exemple:
        
            > pos = motor1.stat.get_node('pos_actual')
            Is equivalent to
            > pos = motor1.stat.pos_actual
            > pos.get()
        """
        try:
            vNode = self.__dict__[name]
        except KeyError:
            try:
                uaNode = self._get_ua_node(name)
            
            except KeyError:
                # it can be an alias
                try:
                    vNode = object.__getattribute__(self, name)
                except AttributeError:
                    raise KeyError('Unknown node %r'%name)
            
            else:
                vNode = self.Node(kjoin(self.key, name), uaNode, vtype=self._get_vtype(name), sid=self._ua_client.server_url.netloc)
                # cash the object inside __dict__
                self.__dict__[name] = vNode
                return vNode
        
        return vNode

class RpcInterface:
    """ Interface between opc-ua nodes of rpc method and pydevmgr set of :class:`RpcNodes` 
    
    Args:
        
        key (str): a unique key defining the interface. This is generally the same key as the host 
                  :class:`Device`. e.g. mgr.motor1.key ==  mgr.motor1.rpc.key
                  key for rpc method interface is not as relevent as for :class:`Interface`
        ua_client (opcua.Client): opcua client connected to some servers
        map (dict): a dictionnary of name/rpc_node pairs 
            Like:     
            
                ::
                  
                    {   'rpcInit'    : 'RPC_Init',
                        'rpcEnable'  : 'RPC_Enable',
                        'rpcDisable' : 'RPC_Disable',
                    # etc ...
                    }
                    
        ua_prefix (str): The prefix for the OPC-UA path like for instance 'MAIN.Motor1'
        ua_namespace (int,str, optional): The ua name space default is 4
    """
    _ua_client = None
    _map = None
    _ua_prefix = None
    _prefix = None
    _name = None
    
    RpcNode = RpcNode
    RPC_ERROR = RPC_ERROR
    
    def __init__(self, key, ua_client, map, ua_prefix,  ua_namespace=DEFAULT_NAME_SPACE):
        self._ua_client = ua_client
        self._map = map

        self._ua_prefix = ua_prefix
        self._key = key
        
        self._ua_namespace = ua_namespace

    def __getattr__(self, attr):
        try:
            return self.__dict__[attr]
        except KeyError:
            return self.get_method(attr)

    @classmethod
    def _get_atypes(cl, key):
        """ look in class __annotations__ to find any parser for a given key """

        for subcl in cl.__mro__:
            try:
                return subcl.__annotations__[key]
            except (KeyError,AttributeError):
                continue
        return None
    
    def clear(self):
        """ clear all cashed data """
        for k,v in list(self.__dict__.items()):
            if isinstance(v, RpcNode):
                self.__dict__.pop(k)
                
    @property
    def key(self):
        return self._key
    
    @property
    def prefix(self):
        return ksplit(self._key)[0]

    
    @property
    def name(self):
        return ksplit(self._key)[1]    
    
    @property
    def map(self):
        return self._map

    def _get_ua_id(self):
        return "ns={};s={}".format(self._ua_namespace, self._ua_prefix)

    def _get_method_id(self, name):
        try:
            ua_name = self._map[name]
        except KeyError:
            raise KeyError('Unknown method %r'%name)
        return "{}:{}".format(self._ua_namespace, ua_name)


    def _get_ua_node(self):
        uaNode = self._ua_client.get_node(self._get_ua_id())
        return uaNode

    def get_method(self, name):
        """ get a rpc method object :class:`RpcNode`
        
        Args:
            name (str): name of the rpc method
        
        !!! Warning if the given name startwith '_' this method may return something which is not a :class:`RpcNode`. No check is done.    
        """
        try:
            m = self.__dict__[name]
        except KeyError:
            m = self.RpcNode( kjoin(self.key, name), self._get_ua_node(), self._get_method_id(name), atypes=self._get_atypes(name))
            self.__dict__[name] = m
        return m



import time
from .tools import BaseSequence, download

    
class BaseSequence:
    counter = 0
    last_cycle_time = 0.0

    def __init__(self):
        self.data = {}
        self.counter = 0
        self.last_cycle_time = 0
        self.seq_state = 0
        
    def update(self):
        raise NotImplementedError('update')
    def __next__(self):
        if not self.next_cycle():
            raise StopIteration()


    def __iter__(self):
        return self

    def next_cycle(self):
        s_time = time.time()
        r = self.run_cycle()
        self.last_cycle_time = time.time()-s_time
        return r

    def run_cycle(self):
        raise NotImplementedError('next_cycle')

    def runner(self, cycle_time=0.1):
        def run(seq=self, cycle_time=cycle_time):
            while seq.next_cycle():
                time.sleep(max( cycle_time-seq.last_cycle_time, 0.0 ))
        return run
        
    def reset(self):
        raise NotImplementedError('reset')


class MoveTwoPosSequence(BaseSequence):
    """ Sequencer that move motor in two positions successively

    At each position a callback can be executed.
    Of course the sequencer is not suitable for time critical application.

    One can simply iter on the object to execute cycles:

        seq = MoveTwoPosSequence(motor, 0.0, 2.0, velocity=0.5, sleep_time=3.0)
        while seq.next_cycle():
            time.sleep(0.1) # this set the min time resolution of the sequencer

    In the exemple above the cycle time of the sequencer is NOT deterministic as
    the time taken by OPC-UA request or the callback can vary. But this should be
    ok in most AIT test situation.

    We can get closer to a deterministic cycle time by doing:

        while seq.next_cycle():
            time.sleep( max(0.1-seq.last_cycle_time, 0.0) )

    Or the exemple above is equivalent on doing:

        > seq.runner(0.1)()

    One can encapsulate the sequencer in a Thread or async func (while it is not an async object)
    The runner method is creating a function ready to be executed in a thread :
    It take one argument, the cycle time:

        seq = MoveTwoPosSequence(motor, 0.0, 2.0, velocity=0.5, sleep_time=3.0)
        t = Thread(target=seq.runner(0.1))
        t.start()
    
    """
    class SEQ_STATE:
        IDL, WAITING_H, WAITING_L, MOVING_L, MOVING_H = range(5)
        STOP = 100

    def __init__(self, motor, l_pos=0.0, h_pos=1.0, velocity=0.1, l_velocity=None,
                 h_velocity=None, sleep_time=0.0, callback=None, h_callback=None, l_callback=None,
                 n_sequence=None
                 ):
        super().__init__()
        self.motor = motor
        self.l_pos = l_pos
        self.h_pos = h_pos
        self.l_velocity = velocity if l_velocity is None else l_velocity
        self.h_velocity = velocity if h_velocity is None else h_velocity

        self.sleep_time = sleep_time

        self.callback = callback
        self.h_callback = callback if h_callback is None else h_callback
        self.l_callback = callback if l_callback is None else l_callback
        self.n_sequence = n_sequence
        
        self.seq_state = self.SEQ_STATE.IDL
        self._action_time = time.time()

    def stop(self):
        self.seq_state = self.SEQ_STATE.STOP

    def reset(self):
        self.seq_state = self.SEQ_STATE.IDL

    def update(self):
        download([self.motor.stat.substate, self.motor.stat.pos_actual], self.data)

    def run_cycle(self):
        S = self.SEQ_STATE
        if self.seq_state == S.STOP:
            self.motor.stop()
            return False


        if self.seq_state in (S.WAITING_H, S.WAITING_L):
            if (time.time()-self._action_time)>self.sleep_time:
                if self.seq_state == S.WAITING_H:

                    self.motor.move_abs(self.l_pos, self.l_velocity)
                    self.seq_state = S.MOVING_L
                else:
                    self.motor.move_abs(self.h_pos, self.h_velocity)
                    self.seq_state = S.MOVING_H
            else:
                return True # need to wait

        elif self.seq_state in (S.MOVING_L, S.MOVING_H):
            self.update()
            if self.data[self.motor.stat.substate.key] == self.motor.SUBSTATE.OP_STANDSTILL:
                if self.seq_state == S.MOVING_H:
                    self.seq_state = S.WAITING_H
                    self._action_time = time.time()
                    if self.h_callback: self.h_callback(self)
                else:
                    self.seq_state = S.WAITING_L
                    self.counter += 1 # counter increase when reaching the low value
                    self._action_time = time.time()
                    if self.h_callback: self.l_callback(self)
                    if self.n_sequence is not None and self.counter>=self.n_sequence:
                        return False

            else:
                return True

        elif self.seq_state == S.IDL:
            self.motor.move_abs(self.l_pos, self.l_velocity)
            self.seq_state = S.MOVING_L
        else:
            raise Exception("BUG seq_state", self.seq_state)

        return True



from .device import Device, _name_attr, GROUP
from . import trk
from .motor import Motor
from . import keys as K
from .interface import Int16, Int32, nodealiasproperty, kjoin
from .manager import Manager # see at the end of file
from .tools import upload

#                      _              _   
#   ___ ___  _ __  ___| |_ __ _ _ __ | |_ 
#  / __/ _ \| '_ \/ __| __/ _` | '_ \| __|
# | (_| (_) | | | \__ \ || (_| | | | | |_ 
#  \___\___/|_| |_|___/\__\__,_|_| |_|\__|
# 

##### ############
# SUBSTATE
@_name_attr
class SUBSTATE(trk.SUBSTATE):
    pass


### ##############
# RPC error
@_name_attr
class RPC_ERROR(trk.RPC_ERROR):
    txt = {
        # add more rpc error description here if necessary @_name_attr takes care of the ones in master class
    }

### ##############
# errors
@_name_attr
class ERROR(trk.ERROR):
    txt = {
        # add more  error description here if necessary @_name_attr takes care of the ones in master class
    }


### ##############
# errors
@_name_attr
class AXIS:
    ALL_AXIS = 0
    AXIS1 = 1
    AXIS2 = 2



### ############# 
# Mode 
@_name_attr
class MODE:
    ENG   = 0 
    OFF   = 1
    AUTO  = 2
    group = {
        ENG		: GROUP.ENG,
        OFF	    : GROUP.STATIC,
        AUTO	: GROUP.TRACKING,
    }
#  _       _             __                
# (_)_ __ | |_ ___ _ __ / _| __ _  ___ ___ 
# | | '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \
# | | | | | ||  __/ |  |  _| (_| | (_|  __/
# |_|_| |_|\__\___|_|  |_|  \__,_|\___\___|

class AdcStatInterface(Device.StatInterface):
    ERROR = ERROR
    MODE = MODE
    SUBSTATE = SUBSTATE
    @nodealiasproperty("track_mode_txt", ["track_mode"])
    def track_mode_txt(self, track_mode):
        return self.MODE.txt.get(track_mode, "UNKNOWN")
        
    
    
    
class AdcCfgInterface(Device.CfgInterface):
    # we can define the type to parse value directly on the class by annotation
    trk_period : Int32
    mot1_signoff:  Int32    
    mot2_signoff:  Int32    
    mot1_signauto: Int32    
    mot2_signauto: Int32    
    mot1_signphi:  Int32    
    mot2_signphi:  Int32    

# redefine the Method interface to include the proper description of the RpcError
class AdcRpcNode(Device.RpcInterface.RpcNode):
    RPC_ERROR = RPC_ERROR

class AdcRpcInterface(Device.RpcInterface):
    RpcNode = AdcRpcNode
    RPC_ERROR = RPC_ERROR
    ##
    # the type of rpcMethod argument can be defined by annotation
    # All method args types must be defined in a tuple
    rpcMoveAbs :  (Int16, float, float)
    rpcMoveRel :  (Int16, float, float)
    rpcMoveAngle: (float,)
    rpcMoveVel:   (Int16, float,)
    rpcStartTrack : (float,)
    
#      _            _          
#   __| | _____   _(_) ___ ___ 
#  / _` |/ _ \ \ / / |/ __/ _ \
# | (_| |  __/\ V /| | (_|  __/
#  \__,_|\___| \_/ |_|\___\___|
#
class Adc(Device,trk.Trk):
    SUBSTATE = SUBSTATE
    MODE = MODE 
    ERROR = ERROR 
    
    RpcInterface  = AdcRpcInterface
    StatInterface = AdcStatInterface
    CfgInterface  = AdcCfgInterface
    
    def __init__(self, *args, **kwargs):
        super(Adc, self).__init__(*args, **kwargs)
        
        axes = self._config['ctrl_config'].get('axes', [])
        motors = []
        for axis in axes:
            axis_conf = self._config[axis]
            name = axis_conf.get('prefix', None)
            motors.append(Motor.from_config(axis_conf['cfgfile'], axis, key=kjoin(self.key, name)) )
        self.motors = motors    
    
    @property
    def motor1(self):
        return self.motors[0]
    
    @property
    def motor2(self):
        return self.motors[1]
    
    def connect(self):
        """ Connect all opc-ua client to servers """
        super(Adc, self).connect()
        for m in self.motors:
            m.connect()
    
    def disconnect(self):
        """ Disconnect client frmo their servers """
        super(Adc, self).disconnect()
        for m in self.motors:
            m.disconnect()
        
    def get_configuration(self, **kwargs):
        """  return a node/value pair dictionary ready to be uploaded 
        
        The node/value dictionary represent the device configuration. 
        
        Args:
            **kwargs : name/value pairs pointing to cfg.name node
                      This allow to change configuration on the fly
                      without changing the config file. 
        """
        cfg_dict = {}
        for m in self.motors:
            cfg_dict.update( m.get_configuration() )
        
        config = self._config 
        
        ctrl_config = config.get(K.CTRL_CONFIG, {})  
        # just update what is in ctrl_config, except axes      
        cfg_dict.update( {self.cfg.get_node(k):v for k,v in ctrl_config.items() if k not in ["axes"]} ) 
        cfg_dict.update( {self.cfg.get_node(k):v for k,v in  kwargs.items() } )
        return cfg_dict
    
    def stop(self):
        """ Stop all ADC motions """
        self.rpc.rpcStop.rcall()
    
    def start_track(self, angle=0.0):
        """ Start tracking (AUTO mode)
        
        Args:
            angle (float, optional): target angle default = 0.0
        """
        self.rpc.rpcStartTrack.rcall(angle)
    
    def move_angle(self, angle=0.0):
        """ Move to angle  (OFF mode)
        
        Args:
            angle (float, optional): target angle default = 0.0
        """
        self.rpc.rpcMoveAngle.rcall(angle)
    
    def move_abs(self, axis, pos, vel):
        """ Move one or all motor to an absolute  position 
        
        Args:
            axis (int): 0 for all motors 1 for axis 1 and 2 for axis 2
            pos (float): target absolute position 
            vel (float): target velocity 
        """
        self.rpc.rpcMoveAbs.rcall(axis, pos, vel)
    
    def move_rel(self, axis, pos, vel):
        """ Move one or all motor to an relative position 
        
        Args:
            axis (int): 0 for all motors 1 for axis 1 and 2 for axis 2
            pos (float): target relative position 
            vel (float): target velocity 
        """
        self.rpc.rpcMoveRel.rcall(axis, pos, vel)
    

    def move_vel(self, axis, vel):
        """ Move one or all motor in velocity 
        
        Args:
            axis (int): 0 for all motors 1 for axis 1 and 2 for axis 2
            vel (float): target velocity 
        """
        self.rpc.rpcMoveVel.rcall(axis, vel)
    

    
Manager.record_new_device_type('Adc', Adc)


Manager.record_cfg_template('Adc', """{name}:
  type: Adc
  interface: Softing
  identifier: PLC1                            # OPCUA Object Identifier
  prefix: MAIN_FAST.adc                       # TO BE UPDATED OPCUA attribute prefix
  simulated: false
  ignored: false
  address: {address}
  simaddr: opc.tcp://134.171.57.209:4840       # Simulation address
  mapfile: "{cfg_dir}/mapAdc.yml"      # TO BE UPDATED
  fits_prefix: "ADC1"
  ctrl_config:
    latitude:              -0.429833092
    longitude:             1.228800386
    axes: ['adc1_motor1', 'adc1_motor2']

  adc1_motor1:
      prefix: "motor1"
      cfgfile: "{cfg_dir}/adc1Motor1.yml" # TO BE UPDATED
  adc1_motor2:
      prefix: "motor2"
      cfgfile: "{cfg_dir}/adc1Motor2.yml" # TO BE UPDATED
""")




from threading import Thread as _Thread
import time

class Thread(_Thread):
    """ Thread with a kill and isrunning method """
    _livesignal = True
    def __init__(self):
        self._livesignal = True
        self._lock = False
        super(Thread, self).__init__()
        

    def kill(self):
        self._livesignal = False

    def isrunning(self):
        return self._livesignal



from PyQt5 import QtWidgets, uic
from .io import find_ui
from .generic import record_widget,  method_caller, method_switcher, RpcError
from pydevmgr import Shutter, Prefixed
from .basic import BaseCtrl, BaseLine

class ShutterCtrl(BaseCtrl):
    
    def initUI(self):
        uic.loadUi(find_ui('shutter_ctrl_frame.ui'), self)
        
    def update_device(self, data, shutter):
        super(ShutterCtrl, self).update_device(data, shutter)
        
    
    def link_device(self, downloader, shutter, altname=None):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        shutter (:class:`pydevmgr.Shutter`):  Shutter device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        super(ShutterCtrl, self).link_device(downloader, shutter, altname)
        downloader, token = self.connection # comming from super 
        try:
            self.open.clicked.disconnect()
            self.close.clicked.disconnect()
        except TypeError:
            pass
            
        # add necessary nodes to the downloader state, substate, status, error_code 
        # are handled  by BaseCtrl.link_device
        # stat = shutter.stat
        #downloader.add_node(token)
        
        # Now link the buttons to an action
        open = method_caller(shutter.open, 
                                 [],
                                 self.feedback
                                 )
        self.open.clicked.connect(open)
        close = method_caller(shutter.close, 
                                 [],
                                 self.feedback
                                 )
        self.close.clicked.connect(close)
        
class ShutterLine(BaseLine):

    def initUI(self):
        uic.loadUi(find_ui('shutter_line_frame.ui'), self)
        
    def update_device(self, data, shutter):
        super(ShutterLine, self).update_device(data, shutter)
        # nothing more 
    
    def link_device(self, downloader, shutter, altname=None,more_methods=[]):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        shutter (:class:`pydevmgr.Shutter`):  Shutter device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        more_methods = [
                        ("OPEN" , shutter.open ,[]), 
                        ("CLOSE", shutter.close,[])
                        ]+list(more_methods)
        super(ShutterLine, self).link_device(downloader, shutter, altname, more_methods = more_methods)
        
        token = self.token # comming from super 
        
                

record_widget("ctrl", "Shutter", ShutterCtrl)
record_widget("line", "Shutter", ShutterLine)


from PyQt5 import QtWidgets, uic
from .io import find_ui
from .generic import record_widget, STYLE, method_caller, method_switcher, RpcError, get_style
from pydevmgr import Motor, Prefixed
from pydevmgr.device import  _name_attr
from .basic import BaseCtrl, BaseLine


class MOVE_MODE:
    ABSOLUTE = 0
    RELATIVE = 1
    VELOCITY = 2
    txt = {
       ABSOLUTE : "ABSOLUTE", 
       RELATIVE : "RELATIVE", 
       VELOCITY : "VELOCITY"
    }
    count = 3

class MotorCtrl(BaseCtrl):
    
    def initUI(self):
        uic.loadUi(find_ui('motor_ctrl_frame.ui'), self)
        
    def update_device(self, data, motor):
        super(MotorCtrl, self).update_device(data, motor)
        
        if self.move_mode.currentIndex()==MOVE_MODE.VELOCITY: # VELOCITY
            self.input_pos_target.setEnabled(False)
        else:
            self.input_pos_target.setEnabled(True)
        
        
        self.pos_actual.setText( "{:.3f}".format(data['pos_actual']) )
        self.pos_error.setText(  "{:.3E}".format(data['pos_error'])  )
        self.pos_target.setText( "{:.3f}".format(data['pos_target']) )
        self.vel_actual.setText( "{:.3f}".format(data['vel_actual']) )
        #self.vel_target.setText( "{:.3f}".format(data['vel_target']) )
        
        self.posname.setText("{}".format(data['pos_name']))
    
    def link_device(self, downloader, motor, altname=None):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        motor (:class:`pydevmgr.Motor`):  Motor device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        super(MotorCtrl, self).link_device(downloader, motor, altname)
        token = self.token # comming from super 
        try:
            self.move.clicked.disconnect()
            self.move_by_posname.currentIndexChanged.disconnect()
        except TypeError:
            pass
        
        # init some field to what the state is 
        
        if motor.is_connected():
            self.input_velocity.setText("%.3f"%motor.cfg.velocity.get())
            self.input_pos_target.setText("%.3f"%motor.stat.pos_actual.get())            
        else:
            # put some dummy values
            self.input_velocity.setText("1.00")
            self.input_pos_target.setText("0.00")            
        
        # update the positoin name menu 
        self.move_by_posname.clear()
        self.move_by_posname.addItems(['']+list(motor.posnames))
        
        self.move_mode.clear()
        self.move_mode.addItems(MOVE_MODE.txt[i] for i in range(MOVE_MODE.count))
        
        # add necessary nodes to the downloader state, substate, status, error_code 
        # are handled  by BaseCtrl.link_device
        stat = motor.stat
        downloader.add_node(token, 
                         stat.pos_actual, stat.pos_error, stat.pos_name, 
                         stat.pos_target, stat.vel_actual,# stat.vel_target,
                        
                        )
        
        def move(idx, pos, vel):
            if idx == MOVE_MODE.ABSOLUTE:
                motor.move_abs(pos, vel)
            elif idx == MOVE_MODE.RELATIVE:
                motor.move_rel(pos, vel)
            elif idx == MOVE_MODE.VELOCITY:
                motor.move_vel(vel)
        
        # Now link the buttons to an action
        move = method_caller(move, 
                                 [self.move_mode.currentIndex, 
                                  self.input_pos_target.text, 
                                  self.input_velocity.text],
                                  self.feedback
                            )
        self.move.clicked.connect(move)
                
        stop = method_caller(motor.stop, [], self.feedback)
        self.stop.clicked.connect(stop)
        
        def move_name(i):
            target_name = self.move_by_posname.currentText()
            if target_name=="":
                return
            self.rpc_feedback.setStyleSheet(get_style(STYLE.NORMAL))
            try:
                motor.move_name(target_name, self.input_velocity.text())
            except (TypeError,ValueError,RpcError) as e: 
                self.rpc_feedback.setText(str(e))
                self.rpc_feedback.setStyleSheet(get_style(STYLE.ERROR))
            self.move_by_posname.setCurrentIndex(0)
            
        self.move_by_posname.currentIndexChanged.connect(move_name)
        

class MotorLine(BaseLine):
        
    def initUI(self):
        uic.loadUi(find_ui('motor_line_frame.ui'), self)
        
    def update_device(self, data, motor):
        super(MotorLine, self).update_device(data, motor)
        self.input_pos_target.update()
        self.input_velocity.update() 
        
        
        pos = data['pos_name'] if data['pos_name'] else "{:.3f}".format(data['pos_actual'])
            
        self.pos_actual.setText( pos )
    
    
    def link_device(self, downloader, motor, altname=None, more_methods=[]):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        motor (:class:`pydevmgr.Motor`):  Motor device
        altname (string, optional): Alternative printed name for the device
        """
        ## disconnect all the button if they where already connected
        
        methods = [
            ("MOVE ABS", motor.move_abs,[self.input_pos_target.text, self.input_velocity.text]),
            ("MOVE REL", motor.move_rel,[self.input_pos_target.text, self.input_velocity.text]), 
            "---"
        ]
        methods += [(name, motor.move_name, [lambda name=name:name,self.input_velocity.text]) for name in motor.posnames]
        methods += list(more_methods)
        super(MotorLine, self).link_device(downloader, motor, altname, more_methods=methods)
        token = self.token # comming from super 
        
        if motor.is_connected():
            self.input_velocity.setText("%.3f"%motor.cfg.velocity.get())
            self.input_pos_target.setText("%.3f"%motor.stat.pos_actual.get())            
        else:
            # put some dummy values
            self.input_velocity.setText("1.00")
            self.input_pos_target.setText("0.00")            
        
        
        stat = motor.stat
        downloader.add_node(token, 
                         stat.pos_actual, stat.pos_name, 
                        )
        




record_widget("ctrl", "Motor", MotorCtrl)
record_widget("line", "Motor", MotorLine)




from pydevmgr import Device
from .io import find_ui
from .generic import STYLE, method_switcher, record_widget, method_setup, get_style
from PyQt5 import QtWidgets, uic


class _Shared:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initUI()
        self.connection  = None
        self._did_failed = False
        self.device_key = ""
    
    def connect_downloader(self, downloader):
        ## create a new connection 
        token = downloader.new_connection()
        self.connection = (downloader, token)
        return token
    
    def disconnect_downloader(self):
        if self.connection:
            previous_downloader, token = self.connection 
            previous_downloader.disconnect(token)
            
    def __del__(self):
        self.disconnect_downloader()
        
    @property
    def token(self):
        if self.connection is None:
            return None
        _, token = self.connection
        return token            

class BaseCtrl(QtWidgets.QFrame, _Shared):
            
    def initUI(self):
        uic.loadUi(find_ui('base_ctrl_frame.ui'), self)
        
        
    def feedback(self, er, txt):
        self.rpc_feedback.setText(txt)
        if er:
            self.rpc_feedback.setStyleSheet(get_style(STYLE.ERROR))
        else:
            self.rpc_feedback.setStyleSheet(get_style(STYLE.NORMAL))
        
    def update_failure(self, er):
        """ executed when the Downloader failed. Must be connected """
        style = get_style(STYLE.UNKNOWN)
        self.setEnabled(False)
        for wgt in [self.state, self.substate, self.error_txt]:
            wgt.setStyleSheet(style)
        self._did_failed = True    
    
    def link_failure(self):
        """ link a failure callback to the current connection """
        ## diconnect the downloader 
        if self.connection is None:
            raise ValueError("cannot link failure, widget is not connected")
        downloader, token = self.connection
        downloader.add_failure_callback(token, self.update_failure)
    
    
    def update_device(self, data, device):
        if self._did_failed:
            self.setEnabled(True)
            style = get_style(STYLE.NORMAL)
            for wgt in [self.state, self.substate, self.error_txt]:
                wgt.setStyleSheet(style)
            self._did_failed = False
        
        self.rpc_feedback.update()
        state = data['state']
        self.state.setText("{}: {}".format(state, data['state_txt']))
        self.state.setStyleSheet(get_style(data['state_group']))
        
        substate = data['substate']
        self.substate.setText("{}: {}".format(substate, data['substate_txt']))
        self.substate.setStyleSheet(get_style(data['substate_group']))
        
        erc = data['error_code']
        self.error_txt.setText("{}: {}".format(erc, data['error_txt']))
        self.error_txt.setStyleSheet( get_style(STYLE.ERROR) if erc else get_style(STYLE.OK) )
    
     
        
    def link_device(self, downloader, device, altname=None, more_methods=[]):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        device (:class:`pydevmgr.Motor`):  device to be linked
        altname (string, optional): Alternative printed name for the device
        """
        try:
            self.state_action.currentIndexChanged.disconnect()
        except TypeError:
            pass
        
        
        ## disconnect all the button if they where already connected
        ## diconnect a previous connection
        self.disconnect_downloader()
        token = self.connect_downloader(downloader) 
        self.device_key = device.key 
        
        stat_methods = [
           ("",       None,          []), 
           ("INIT",   device.init,   []), 
           ("ENABLE", device.enable, []),
           ("DISABLE",device.disable,[]),
           ("RESET",  device.reset,  []) 
        ]+list(more_methods)
        
        # action menu 
        self.state_action.clear()
        # Add items to combo and remove string "---" separator from stat_methods
        stat_methods = method_setup(self.state_action, stat_methods)
                
        self.name.setText( device.key if altname is None else altname )
        
        stat = device.stat
        downloader.add_node(token, 
                         stat.state, stat.state_txt, stat.state_group, 
                         stat.substate, stat.substate_txt, stat.substate_group,  
                         stat.error_code, stat.error_txt,                                      
                        )
        # Add the callback to the downloader            
        def update(data):
            return self.update_device(downloader.get_data(device.key), device)                        
        downloader.add_callback(token, update)
        
        state_action = method_switcher(stat_methods, self.feedback, 
                                       lambda : self.state_action.setCurrentIndex(0)
                                       )
        self.state_action.currentIndexChanged.connect(state_action)



class BaseLine(QtWidgets.QFrame, _Shared):
    
    
    
        
    def initUI(self):
        uic.loadUi(find_ui('base_line_frame.ui'), self)
    
    def feedback(self, er, txt):
        if er:
            #self.state_action.setStyleSheet("border-color: red;")
            self.state_action.setItemText(0, '!ERROR')
        else:
            self.state_action.setItemText(0, '')
            #self.state_action.setStyleSheet(STYLE.NORMAL)
    
    def update_failure(self, er):
        """ executed when the Downloader failed. Must be connected """
        style = get_style(STYLE.UNKNOWN)
        self.setEnabled(False)
        for wgt in [self.substate]:
            wgt.setStyleSheet(style)
        self._did_failed = True    
    
    def link_failure(self):
        """ link a failure callback to the current connection """
        ## diconnect the downloader 
        if self.connection is None:
            raise ValueError("cannot link failure, widget is not connected")
        downloader, token = self.connection
        downloader.add_failure_callback(token, self.update_failure)
    
    def update_device(self, data, device):
        if self._did_failed:
            self.setEnabled(True)
            style = get_style(STYLE.NORMAL)
            for wgt in [self.substate]:
                wgt.setStyleSheet(style)
            self._did_failed = False
        
        substate = data['substate']
        self.substate.setText("{}".format(data['substate_txt']))
        self.substate.setStyleSheet(get_style(data['substate_group']))
    
    def link_device(self, downloader, device, altname=None, more_methods=[]):
        """ Link a device to the widget 
        
        downloader (:class:`pydevmgr.Downloader`): a Downloader object 
        device (:class:`pydevmgr.Motor`):  device to be linked
        altname (string, optional): Alternative printed name for the device
        """
        try:
            self.state_action.currentIndexChanged.disconnect()
        except TypeError:
            pass
        
        ## disconnect all the button if they where already connected
        ## diconnect a previous connection
        self.disconnect_downloader()
        token = self.connect_downloader(downloader) 
        self.device_key = device.key 
        
        
        stat_methods = [
           ("",       None,          []), 
           ("INIT",   device.init,   []), 
           ("ENABLE", device.enable, []),
           ("DISABLE",device.disable,[]),
           ("RESET",  device.reset,  []) 
        ]+list(more_methods)
        
        # action menu 
        self.state_action.clear()
        # Add items to combo and remove string "---" separator from stat_methods
        stat_methods = method_setup(self.state_action, stat_methods)
        
        self.name.setText( device.key if altname is None else altname )
        
        stat = device.stat
        downloader.add_node(token, 
                         stat.substate, stat.substate_txt, stat.substate_group                                    
                        )
        # Add the callback to the downloader            
        def update(data):
            return self.update_device(downloader.get_data(device.key), device)                        
        downloader.add_callback(token, update)
        
        state_action = method_switcher(stat_methods, self.feedback, 
                                       lambda : self.state_action.setCurrentIndex(0)
                                       )
        self.state_action.currentIndexChanged.connect(state_action)
        



record_widget("ctrl", "Device", BaseCtrl)
record_widget("line", "Device", BaseLine)
  


from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QApplication, QMainWindow
import sys
from pydevmgr import Motor, Downloader
from pydevmgr_qt.motor import MotorCtrl, BaseCtrl

class Test(QtWidgets.QWidget):
    def __init__(self, *args):
        super(Test, self).__init__(*args)
        self.label = QtWidgets.QLabel(self)
        self.label.setText('Hello')
class Test2(QtWidgets.QWidget):
    def __init__(self, *args):
        super(Test2, self).__init__(*args)
        self.label = QtWidgets.QLabel(self)
        self.label.setText('      ----Hello')
        
motor1, motor2 = None, None
def main():
    global motor1, motor2
    #motor = Motor.from_config('fcf/devmgr/server/m403.yml', 'm403')
    motor1 = Motor.from_config('fcf/devmgr/server/motor1.yml', 'motor1')
    motor2 = Motor.from_config('fcf/devmgr/server/motor2.yml', 'motor2')
    motor1.connect(); motor1.configure()
    motor2.connect(); motor2.configure()
    #motor.configure() # configure true OPC-UA the device
    downloader = Downloader()
    
    app = QApplication(sys.argv)
    win = QtWidgets.QWidget()
    layout =  QtWidgets.QVBoxLayout(win)

    motFrame1 = MotorCtrl(win)
    motFrame1.link_device(downloader, motor1)
    motFrame1.link_failure()
    
    motFrame2 = BaseCtrl(win)
    motFrame2.link_device(downloader, motor2)
    motFrame2.link_failure()
    
    
    connect = QtWidgets.QPushButton("Connect")
    disconnect = QtWidgets.QPushButton("Disconnect")
    connect.clicked.connect( motor1.connect )
    disconnect.clicked.connect( motor1.disconnect )
    
    
    layout.addWidget(motFrame1)
    layout.addWidget(motFrame2)
    
    layout.addWidget(connect)
    layout.addWidget(disconnect)
    
    
    #### test ####
    # test1 = Test(win)
    # test2 = Test2(win)
    #
    # layout.addWidget(test1)
    # layout.addWidget(test2)


    #win.setLayout( layout )
    win.show()
    timer = QtCore.QTimer()
    timer.timeout.connect(downloader.download)
    timer.start(100)
    sys.exit(app.exec_())


if __name__ == '__main__':
    try:
        main()
    finally:
        motor1.disconnect()
        motor2.disconnect()



from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QApplication, QMainWindow
import sys
from pydevmgr import Motor, Downloader, Manager
from pydevmgr_qt.motor import MotorCtrl
from pydevmgr_qt.manager import ManagerDevices

class Test(QtWidgets.QWidget):
    def __init__(self, *args):
        super(Test, self).__init__(*args)
        self.label = QtWidgets.QLabel(self)
        self.label.setText('Hello')
class Test2(QtWidgets.QWidget):
    def __init__(self, *args):
        super(Test2, self).__init__(*args)
        self.label = QtWidgets.QLabel(self)
        self.label.setText('      ----Hello')
        
mgr = None

live = True
def main():
    global mgr
    
    #mgr = Manager.from_config('fcf/devmgr/server/cfgM403.yml', 'fcs')
    #mgr = Manager.from_config('fcf/devmgr/server/tins.yml', 'fcs', extra_file="fcf/devmgr/server/tins_pydevmgr.yml")
    mgr = Manager.from_config('fcf/devmgr/server/tins.yml', 'fcs')
    
    
    if live:
        mgr.connect_all()
        mgr.configure_all() # configure true OPC-UA the devices
    
    downloader = Downloader()


    app = QApplication(sys.argv)
    #app.setStyleSheet(style)
    #app.setStyle("Fusion")
    #win = ManagerMainWindow()
    win = ManagerDevices()
    win.link_device(downloader, mgr, widget_type='ctrl')
    
    # win = QtWidgets.QWidget()
    # layout =  QtWidgets.QVBoxLayout(win)
    #
    # managerWidget = ManagerCtrl(win)
    # managerWidget.link_device(downloader, mgr)
    #
    # layout.addWidget(managerWidget)
    #win.setLayout( layout )
    win.show()
    timer = QtCore.QTimer()
    timer.timeout.connect(downloader.download)
    if live:
        timer.start(200)
    sys.exit(app.exec_())


if __name__ == '__main__':
    try:
        main()
    finally:
        if live and mgr: 
            mgr.disconnect_all()



from pydevmgr import Device, RpcError, GROUP


_enum = -1 
def _inc(i=None):
    """ number increment to use in frontend 
    
    _inc(0) # reset increment to 0 and return 0 
    _inc()  # increment and return incremented number 
    """
    global _enum
    _enum = _enum+1 if i is None else i
    return _enum

# TODO: change this by a real logging system 
error = print

class STYLE(GROUP):
    """ A collection of style IDs derived from GROUPs in pydevmgr + extra stuff """
    NORMAL  = "NORMAL"
    ODD     = "ODD"
    EVEN    = "EVEN"
    ERROR_TXT = "ERROR_TXT"
    OK_TXT =    "OK_TXT"
    
""" Associate STYLE IDs to qt stylSheet """
qt_style_loockup = {
    STYLE.NORMAL  : "background-color: white;",
    STYLE.IDL  : "background-color: white;",
    STYLE.WARNING : "background-color: #ff9966;",
    STYLE.ERROR   : "background-color: #cc3300;",
    STYLE.OK      : "background-color: #99cc33;",
    STYLE.NOK     : "background-color: #ff9966;",
    STYLE.BUZY    : "background-color: #ffcc00;",
    STYLE.UNKNOWN : "",
    STYLE.ODD     : "background-color: #E0E0E0;",
    STYLE.EVEN    : "background-color: #F8F8F8;",
    STYLE.ERROR_TXT : "color: #cc3300;",
    STYLE.OK_TXT : "color: black;",
}

def get_style(style):
    return qt_style_loockup.get(style, "")

""" Associate a state to a style """
state_style_loockup = {
   Device.STATE.NONE  : STYLE.UNKNOWN, 
   Device.STATE.OP    : STYLE.OK, 
   Device.STATE.NOTOP : STYLE.NOK,
}


def method_setup(combo, items):
    """ Does the same thing than combo_box.setItems 
    
    except that input is a list of (name, func, input_func) as for  method_switcher
    
    """
    output= []
    idx = combo.count()
    for item in items:
        if isinstance(item, str):
            if item == "---":
                combo.insertSeparator(idx)
                # inserting a separator increase the index so add a dummy action
                output.append(("", None, []))
                
            else:
                name = item
                item = (name, None, [])
                combo.addItem(name)
                output.append(item)
        else:
            name, _, _ = item
            combo.addItem(name)
            output.append(item)
        idx = combo.count()
    return output 

def method_caller(func, input_funcs, feedback=None):
    """ Build a method to togle a function 
    
    Args:
        func (callable): function to call
        input_funcs (list of callable): each member return an argument for the function
        feedbac: func with signature func(er, txt)
    
    Returns:
        method : a callable obect with signature func()
        
    """
    if feedback:        
        def call_method():
                      
            try:
                func(*(fi() for fi in input_funcs))
            except (RpcError,TypeError,ValueError) as e:
                feedback(True, str(e))
            else:
                feedback(False, '')
    else:
        def call_method():
            try:
                func(*(fi() for fi in input_funcs))
            except (RpcError,TypeError,ValueError) as e:
                error(str(e))    
    return  call_method
    
def method_switcher( func_inputs, feedback=None, reset=None):
    """ Build method that take an index number and trigger the associated method 
    
    
    Args:
        func_inputs: list of (func, inputs), inputs is alist of callable 
                     each member return an argument for the function
        feedbac: (optional): func with signature func(er,txt)
        reset (callable, optional): to execute after each call
    
    Returns:
        method : a callable obect with signature func(i)
        
    """
    if feedback:    
        def call_method(idx):
            _, func, input_funcs = func_inputs[idx]
            if func is None:
                return 
            
            try:
                func(*(fi() for fi in input_funcs))
            except (RpcError,TypeError,ValueError) as e:
                feedback(True, str(e))
            else:
                feedback(False, '')
            if reset: reset()
    else:
        def call_method(idx):
            func, input_funcs = func_inputs[idx]
            if func is None:
                return 
            try:
                func(*(fi() for fi in input_funcs))
            except (RpcError,TypeError,ValueError) as e:
                error(str(e))
            if reset: reset()
    
    return  call_method                           


_pydevmgr_widgets = {}
def record_widget(widget_type, dev_type, constructor):
    """ Record a new global widget 
    
    Args:
        widget_type (string): description of the widget like "ctrl"
        dev_type (:class:`Device`): A device class 
        constructor: A constructor for the widget  
    """
    _pydevmgr_widgets[(widget_type, dev_type)] = constructor

def get_widget_constructor(widget_type, dev_type):
    try:
        return _pydevmgr_widgets[(widget_type, dev_type)]
    except KeyError:
        raise ValueError("Unknown widget of type {} for a {}".format(widget_type, dev_type))




from PyQt5 import QtWidgets, uic
from PyQt5.QtWidgets import QVBoxLayout, QMainWindow
from PyQt5.QtCore import Qt

from .generic import get_widget_constructor, STYLE, get_style 
from .io import find_ui
from .basic import BaseCtrl 

from pydevmgr.device import STATE
from pydevmgr import Prefixed


# TODO change the warning function to something else
warning = print




class ManagerDevices(QtWidgets.QWidget):
    _did_failed = False
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initUI()
        self.devices = {}
        self.connection = None
        self.check_status = {}
        self.dev_widgets = []
        
    def initUI(self):
        uic.loadUi(find_ui('manager_main.ui'), self)
    
    def update_failure(self, er):
        """ executed when the Downloader failed. """
        self.setEnabled(False)
        self._did_failed = True    
    
    def link_failure(self):
        """ link a failure callback to the current connection """
        ## diconnect the downloader 
        if self.connection is None:
            raise ValueError("cannot link failure, widget is not connected")
        downloader, token = self.connection
        downloader.add_failure_callback(token, self.update_failure)
    
    
    
    def update_manager(self, data, mgr):
        if self._did_failed:
            self.setEnabled(True)
            self._did_failed = False
        
        state = data['state']
        
        self.state.setText("{}: {}".format(state, data['state_txt']))
        self.state.setStyleSheet(get_style(data['state_group']))
        
        
        for dev_widget in self.dev_widgets:
            if hasattr(dev_widget, "check"):
                s = self.check_status.get(dev_widget.device_key, True)
                if s != dev_widget.check.isChecked():
                    dev_widget.check.setChecked(s)
                
    def link_device(self, downloader, manager):
        
        if self.connection:
            previous_downloader, token = self.connection
            previous_downloader.disconnect(token)
        ## create a new connection 
        token = downloader.new_connection()
        self.connection = (downloader, token)
        
        
        try:
            self.state_action.currentIndexChanged.disconnect()
            self.layout_selector.disconnect()
            self.check_all.disconnect()
        except TypeError:
            pass
        
        self.state_action.clear()
        self.state_action.addItems(['', "INIT", "ENABLE", "DISABLE", "RESET"])
        self.state.setText("UNKNOWN")
        
        self.layout_selector.clear()

        self.name.setText( manager.key )

        # add the necessary nodes to the downloader
        downloader.add_node(token, manager.stat.state, manager.stat.state_txt, manager.stat.state_group)
        
        def update(data):
            return self.update_manager(downloader.get_data(manager.key), manager) 
        
        downloader.add_callback(token, update)

        def state_action(idx):
            if idx==0:
                return
            erc = 0
            if idx==1:
                for dev in manager.devices():
                    if self.check_status.get(dev.key, True):
                        dev.rpc.rpcInit.call()
            elif idx==2:
                for dev in manager.devices():
                    if self.check_status.get(dev.key, True):
                        dev.rpc.rpcEnable.call()
            elif idx==3:
                for dev in manager.devices():
                    if self.check_status.get(dev.key, True):
                        dev.rpc.rpcDisable.call()
            elif idx==4:
                for dev in manager.devices():
                    if self.check_status.get(dev.key, True):
                        dev.rpc.rpcReset.call()
            self.state_action.setCurrentIndex(0)
        self.state_action.currentIndexChanged.connect(state_action)
        
        extra = manager.extra    
        layouts = extra.get("layouts", [])
        if not layouts: 
            extra = build_standard_layout(manager)
            layouts = extra['layouts']
        
        
        self.layout_selector.clear()
        self.layout_selector.addItems(layouts)
        
        # container_layout = QtWidgets.QVBoxLayout()
        # self.device_container.setLayout(vlayout)
        self.dev_widgets = []
        if layouts:
            ly = layouts[0]
            device_layout = QtWidgets.QVBoxLayout()
            self.device_container.setLayout(device_layout)
            make_layout(device_layout, extra[ly], downloader, manager, self.dev_widgets, self.check_status)
        
        
        def layout_selector(idx):
            #self.device_container.clear()
            for dev_widget in self.dev_widgets:
                dev_widget.disconnect_downloader()
                if hasattr(dev_widget, "check"):
                    dev_widget.check.disconnect() # remove call backs 
                
            for i in reversed(range(device_layout.count())): 
                device_layout.itemAt(i).widget().setParent(None)
            self.dev_widgets.clear()
                
            make_layout(device_layout, extra[layouts[idx]], downloader, manager, self.dev_widgets, self.check_status)
        
        self.layout_selector.currentIndexChanged.connect(layout_selector)
        
        def check_all():
            s = self.check_all.isChecked()
            for dev_widget in self.dev_widgets:
                self.check_status[dev_widget.device_key] = s
        self.check_all.stateChanged.connect(check_all)
        



## ########################################
def build_standard_layout(manager):
    """ Build a standard layout for a given manager 
    
    This is used when no *_extra.yml file was given 
    """
    extra = {}
    devices = [dev.name for dev in manager.devices()]
    extra['layouts'] = ["One Line", "Basic Ctrl"]
    extra['One Line'] = {
          "type": "tabs",
          "children":[{"type":"tab", 
                       "name":"Devices",
                       "children":[{
                           "type"   :"devices",
                           "devices" :devices, 
                           "widget_type": "line"
                           }]
                    }]
    }
    extra['Basic Ctrl'] = {
          "type": "tabs",
          "children":[{"type":"tab", 
                       "name":"Devices",
                       "children":[{
                           "type"    :"devices",
                           "devices" :devices, 
                           "widget_type": "ctrl"
                           }]
                    }]    
    }
    return extra



class ManagerCtrl(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initUI()
        self.connection = None

    def initUI(self):
        uic.loadUi(find_ui('manager_ctrl.ui'), self)
        #self.show()

    def update_device(self, data, mgr):
        state = data['state']
        self.state.setText("{}: {}".format(state, data['state_txt']))
        self.state.setStyleSheet(get_style(data['state_group']))
        
    def link_device(self, downloader, manager):
        
        if self.connection:
            previous_downloader, token = self.connection
            previous_downloader.disconnect(token)
        ## create a new connection 
        token = downloader.new_connection()
        self.connection = (downloader, token)
        
        
        try:
            self.state_action.currentIndexChanged.disconnect()
        except TypeError:
            pass
        
        self.state_action.clear()
        self.state_action.addItems(['', "INIT", "ENABLE", "DISABLE", "RESET"])
        self.state.setText("UNKNOWN")

        # add the necessary nodes to the downloader
        downloader.add_node(token, manager.stat.state, manager.stat.state_txt, manager.stat.state_group)
        
        def update(data):
            return self.update_device(downloader.get_data(manager.key), manager) 
        
        downloader.add_callback(token, update)

        def state_action(idx):
            if idx==0:
                return
            erc = 0
            if idx==1:
                for dev in manager.devices():
                    dev.rpc.rpcInit.call()
            elif idx==2:
                for dev in manager.devices():
                    dev.rpc.rpcEnable.call()
            elif idx==3:
                for dev in manager.devices():
                    dev.rpc.rpcDisable.call()
            elif idx==4:
                for dev in manager.devices():
                    dev.rpc.rpcReset.call()
            self.state_action.setCurrentIndex(0)
        self.state_action.currentIndexChanged.connect(state_action)


class ManagerMainWindow(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initUI()

    def initUI(self):
        self.top = 100
        self.left = 100
        self.width = 600
        self.height = 600
        #self.setGeometry(self.left, self.top, self.width, self.height)
        self.setMinimumSize(600,300)
        self.layout = QVBoxLayout(self)

        self.ctrl = ManagerCtrl()
        #self.ctrl.move(0, 0)

        self.devices = ManagerDevices()

        #self.devices.move(0,0)
        self.layout.addWidget(self.ctrl)
        self.layout.addWidget(self.devices)

        #self.layout.setAlignment(Qt.AlignTop)
        #self.layout.addStretch()
        #self.layout.addStretch()
        #self.setLayout(self.layout)

        #self.show()

    def link_device(self, downloader, manager):
        self.ctrl.link_device(downloader, manager)
        self.devices.link_device(downloader, manager)
        
        
def make_layout(layout, laydef, downloader, manager, dev_widgets, check_status):
    _make_layout_walk(layout, laydef, downloader, manager, 0, dev_widgets, check_status)
    
def _make_layout_walk(layout, lydef, downloader, manager, child_num, dev_widgets, check_status):
    
    tpe = lydef['type']
    if tpe == "vertical":
        
        container = QtWidgets.QWidget()        
        child_layout = QtWidgets.QVBoxLayout()
        container.setLayout(child_layout)
        layout.addWidget(container)
        
        for i, lydef in enumerate(lydef['children']):
             _make_layout_walk(child_layout, lydef, downloader, manager, i, dev_widgets, check_status)
        child_layout.addStretch()
    
    elif tpe == "horizontal":
        
        container = QtWidgets.QWidget()        
        child_layout = QtWidgets.QHBoxLayout()
        container.setLayout(child_layout)
        layout.addWidget(container)
        
        for i, lydef in enumerate(lydef['children']):
             _make_layout_walk(child_layout, lydef, downloader, manager, i, dev_widgets, check_status)
        #child_layout.addStretch()


        
    elif tpe == "tabs":
        
        tabs = QtWidgets.QTabWidget()
        layout.addWidget(tabs)
        for i, lydef in enumerate(lydef['children']):
            if lydef['type'] != "tab":
                raise ValueError("Only tab are allowed inside tabs")
            
            tab = QtWidgets.QScrollArea()
            
            child_container = QtWidgets.QWidget()
            child_layout = QtWidgets.QVBoxLayout()
            child_container.setLayout(child_layout)
            
            tab.setWidgetResizable(True)
            tab.setWidget(child_container)
            
            tabs.addTab(tab, lydef.get('name', 'tab %d'%(i+1)))
            for j, lydef in enumerate(lydef['children']):
                _make_layout_walk(child_layout, lydef, downloader, manager, j, dev_widgets, check_status)
            child_layout.addStretch()    
        
    
    elif tpe == "device":
        device = manager.get_device(lydef['device'])
        try:
            Widget = get_widget_constructor(lydef['widget_type'], device.dev_type)
        except ValueError:
            
            warning("Cannot find a %r widget for device of type %r"%(lydef['widget_type'], device.dev_type))
            Widget =   get_widget_constructor(lydef['widget_type'], 'Device')
        
        widget = Widget()
        layout.addWidget(widget)
        if hasattr(widget, "check"):
            widget.check.show()
            def checkChanged():
                check_status[device.key] = widget.check.isChecked()
            widget.check.stateChanged.connect(checkChanged)    
        
        dev_widgets.append(widget)
        widget.link_device(downloader, device, altname=device.name)
    
    elif tpe == "all_devices":
        devices = manager.devices()
        widget_type = lydef['widget_type']
        for i,device in enumerate(devices):
            nlydef = {**lydef, 
                "type":"device", "device":device.name
            }
            _make_layout_walk(layout, nlydef, downloader, manager, i, dev_widgets, check_status)
    
    elif tpe == "devices":
        devices = lydef['devices']
        widget_type = lydef['widget_type']
        for i,device in enumerate(devices):
            nlydef = {**lydef, 
                      "type":"device", "device":device,
                     }
            _make_layout_walk(layout, nlydef, downloader, manager, i, dev_widgets, check_status)
    
    elif tpe == "device_types":
        devices = manager.devices()
        dev_types = lydef['device_types']
        
        for i,device in enumerate(devices):
            if device.dev_type in dev_types:
                nlydef = {**lydef, 
                    "type":"device", "device":device.name,
                }
                _make_layout_walk(layout, nlydef, downloader, manager, i, dev_widgets, check_status) 
                
    elif tpe == "device_type":  
        nlydef = {**lydef, "type":"device_types", "device_types":[lydef['device_type']]}
        _make_layout_walk(layout, nlydef, downloader, manager, i, dev_widgets, check_status)      
        
    else:
        raise ValueError("Unknown type %r"%tpe)


""" Not a device but define some share methods end constant for tracking devices """
from .device import _name_attr, GROUP
from .motor import Motor
from .tools import upload

##### ############
# SUBSTATE
@_name_attr
class SUBSTATE:
    NONE =  0

    NOTOP_NOTREADY =  100
    NOTOP_READY = 101
    NOTOP_INITIALIZING = 102
    NOTOP_ABORTING = 107
    NOTOP_RESETTING = 109
    NOTOP_ENABLING = 110
    
    NOTOP_ERROR =  199

    OP_DISABLING 	= 205
    OP_STANDSTILL	= 216
    OP_MOVING		= 217
    OP_SETTING_POS	= 218
    OP_STOPPING		= 219
    OP_TRACKING		= 220
    OP_PRESETTING	= 221

    OP_ERROR =299
    group = {
        NOTOP_NOTREADY     :  GROUP.NOK, 
        NOTOP_READY        :  GROUP.NOK,
        NOTOP_INITIALIZING :  GROUP.BUZY, 
        NOTOP_ABORTING     :  GROUP.BUZY, 
        NOTOP_RESETTING    :  GROUP.BUZY, 
        NOTOP_ENABLING     :  GROUP.BUZY, 
        NOTOP_ERROR        :  GROUP.ERROR, 
        OP_DISABLING 	  :  GROUP.BUZY, 
        OP_STANDSTILL	  :  GROUP.OK,
        OP_MOVING		  :  GROUP.BUZY, 
        OP_SETTING_POS	  :  GROUP.BUZY, 
        OP_STOPPING		  :  GROUP.BUZY, 
        OP_TRACKING		  :  GROUP.OK, 
        OP_PRESETTING	  :  GROUP.BUZY, 
        OP_ERROR          :  GROUP.ERROR,   
    }

SUBSTATE.GROUP_BUZY = (SUBSTATE.NOTOP_INITIALIZING, SUBSTATE.OP_MOVING, 
                       SUBSTATE.OP_STOPPING,SUBSTATE.OP_PRESETTING)
SUBSTATE.GROUP_ERROR = (SUBSTATE.NOTOP_ERROR, SUBSTATE.OP_ERROR)
SUBSTATE.GROUP_OK = (SUBSTATE.OP_STANDSTILL,SUBSTATE.OP_TRACKING)



### ##############
# ERROR
@_name_attr
class ERROR(Motor.ERROR):
    # add more error if necessary 
    txt = dict(Motor.ERROR.txt, **{
        # add more error description here if necessary 
    })

### ##############
# RPC error
@_name_attr
class RPC_ERROR(Motor.RpcInterface.RPC_ERROR):
    # add rpc error if necessary
    txt = dict(Motor.RpcInterface.RPC_ERROR.txt, **{
        # add more rpc error description here if necessary
    })

    
class Trk:
    # note start_track not defined here 
    # because specific (Adc.start_track take one arg Drot.start_track two)
        
    def stop_track(self):
        self.rpc.rpcStopTrack.rcall()
    
    def is_moving(self, data=None):
        """ -> True is axis is moving """
        substate =  self.stat.substate.get() if data is None else data[self.stat.substate.key]
        return substate == self.SUBSTATE.OP_MOVING

    def is_standstill(self, data=None):
        """ -> True is axis is standing still """
        substate =  self.stat.substate.get() if data is None else data[self.stat.substate.key]
        return substate == self.SUBSTATE.OP_STANDSTILL
    
    def is_tracking(self, data=None):
        """ -> True if device is tracking """
        substate =  self.stat.substate.get() if data is None else data[self.stat.substate.key]
        return substate == self.SUBSTATE.OP_TRACKING
    
    def is_presetting(self, data=None):
        """ -> True if device is presetting """
        substate = self.stat.substate.get() if data is None else data[self.stat.substate.key]
        return substate == self.SUBSTATE.OP_PRESETTING


from .device import Device, _name_attr, GROUP
from . import keys as K
from .manager import Manager # see at the end of file
from .interface import Int16, Int32
from .tools import upload

#                      _              _   
#   ___ ___  _ __  ___| |_ __ _ _ __ | |_ 
#  / __/ _ \| '_ \/ __| __/ _` | '_ \| __|
# | (_| (_) | | | \__ \ || (_| | | | | |_ 
#  \___\___/|_| |_|___/\__\__,_|_| |_|\__|
# 


##### ############
# SUBSTATE
@_name_attr
class SUBSTATE:
    NONE                  =   0
    NOTOP_NOTREADY		  = 100
    NOTOP_READY		      = 101
    NOTOP_INITIALISING	  = 102
    NOTOP_ERROR			  = 199
    OP_DISABLING		  = 205
    OP_POS                = 203
    OP_AUTO               = 204
    OP_ERROR			  = 299
    group = {
        NONE                   : GROUP.UNKNOWN,
        NOTOP_NOTREADY         : GROUP.NOK,
        NOTOP_READY            : GROUP.NOK,
        NOTOP_INITIALISING     : GROUP.BUZY,
        NOTOP_ERROR            : GROUP.ERROR, 
        OP_DISABLING		   : GROUP.BUZY,
        OP_POS                 : GROUP.OK, 
        OP_AUTO                : GROUP.OK, 
        OP_ERROR               : GROUP.ERROR, 
    }

### ##############
# ERROR
@_name_attr
class ERROR:
    OK		      = 0
    HW_NOT_OP     = 1			
    ON_FAILURE    = 2
    MAXON         = 3
    OUT_OF_RANGE  = 4
    USER2BIT_ZERO = 5
	
	# Simulator errors
    NOT_INITIALISED		= 90
    ZERO_POINTER		= 100	
    txt = {
        OK:				   'OK',
        HW_NOT_OP:		   'ERROR: TwinCAT not in OP state or CouplerState not mapped.',
        ON_FAILURE:        'ERROR: Piezo HW failure',
        MAXON:             'ERROR: Maximum ON time exceeded.',
        OUT_OF_RANGE:      'ERROR: Piezo set position out of range',
        USER2BIT_ZERO:     'ERROR: cfg.lrUser2Bit has zero value.',
    }

### ##############
# RPC error
@_name_attr
class RPC_ERROR:
    OK			                =  0
    NOT_OP                      = -1
    NOT_NOTOP_READY		        = -2
    NOT_NOTOP_NOTREADY          = -3
    MOVING_USER                 = -5
    MOVING_BIT                  = -6
    LOCAL                       = -7
    txt = {
    	OK:						 'OK',
        NOT_OP:					 'Cannot control device. Not in OP state.',
        NOT_NOTOP_READY:		 'Call failed. Not in NOTOP_READY.',
        NOT_NOTOP_NOTREADY:		 'Call failed. Not in NOTOP_NOTREADY/ERROR.',
        LOCAL:					 'RPC calls not allowed in Local mode.',
        MOVING_USER:			 'Set user value out of range.',
        MOVING_BIT:				 'Set bit value out of range.',
    }


#  _       _             __                
# (_)_ __ | |_ ___ _ __ / _| __ _  ___ ___ 
# | | '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \
# | | | | | ||  __/ |  |  _| (_| | (_|  __/
# |_|_| |_|\__\___|_|  |_|  \__,_|\___\___|

class PiezoStatInterface(Device.StatInterface):
    # keys used by default by update_to
    ERROR = ERROR
    SUBSTATE = SUBSTATE

class PiezoCfgInterface(Device.CfgInterface):
    # we can define the type to parse value directly on the class by annotation
    max_on : Int32 
    num_axis : Int16
    full_range1: Int16  
    full_range2: Int16  
    full_range3: Int16
    home1: Int16  
    home2: Int16  
    home3: Int16
    lower_limit1: Int16  
    lower_limit2: Int16  
    lower_limit3: Int16
    upper_limit1: Int16  
    upper_limit2: Int16  
    upper_limit3: Int16
    user_offset_input1: Int16  
    user_offset_input2: Int16  
    user_offset_input3: Int16
    user_offset_output1: Int16  
    user_offset_output2: Int16  
    user_offset_output3: Int16
    

# redefine the Method interface to include the proper description of the RpcError
class PiezoRpcNode(Device.RpcInterface.RpcNode):
    RPC_ERROR = RPC_ERROR

class PiezoRpcInterface(Device.RpcInterface):
    RpcNode = PiezoRpcNode
    RPC_ERROR = RPC_ERROR
    ##
    # the type of rpcMethod argument can be defined by annotation
    # All args types must be defined in a tuple
    
    rpcMoveBits : (Int16,)*3
    rpcMoveUser : (float,)*3


#      _            _          
#   __| | _____   _(_) ___ ___ 
#  / _` |/ _ \ \ / / |/ __/ _ \
# | (_| |  __/\ V /| | (_|  __/
#  \__,_|\___| \_/ |_|\___\___|
#
class Piezo(Device):
    SUBSTATE = SUBSTATE
    ERROR = ERROR

    RpcInterface = PiezoRpcInterface
    StatInterface = PiezoStatInterface
    CfgInterface = PiezoCfgInterface

         
    def auto(self):
        """ turn on auto mode """
        self.rpc.rpcAuto.rcall()   
    
    def home(self):
        """ send  piezos home  """
        self.rpc.rpcHome.rcall()
    
    def move_bits(self, pos1=0, pos2=0, pos3=0):
        """ move piezos to bits position 
        
        Args:
            pos1 (int): piezo 1 position (bits)
            pos2 (int): piezo 2 position (bits) 
            pos3 (int): piezo 3 position (bits)
        """
        # pos1, pos2, pos3 are piezo set positions in bits - integers.
        self.rpc.rpcMoveBits.rcall(pos1, pos2, pos3)
    
    def move_user(self, pos1=0.0, pos2=0.0, pos3=0.0):
        """ move piezos to user  position 
        
        Args:
            pos1 (float): piezo 1 position (user)
            pos2 (float): piezo 2 position (user) 
            pos3 (float): piezo 3 position (user)
        """
        # pos1, pos2, pos3 are piezo set positions in UU - float.
        self.rpc.rpcMoveUser.rcall(pos1, pos2, pos3)
    
    def pos(self):
        self.rpc.rpcPos.rcall()
    
    def stop(self):
        """ stop movement """
        self.rpc.rpcStop.rcall()
    

Manager.record_new_device_type('Piezo', Piezo)

Manager.record_cfg_template('Piezo', """{name}:
  type: Piezo 
  interface: Softing
  identifier: PLC1                             # OPCUA Object Identifier
  prefix: MAIN.Piezo1                        # TO BE UPDATED OPCUA attribute prefix
  simulated: false
  address: {address}     # TO BE UPDATED OPCUA server address 
  simaddr: opc.tcp://134.171.2.213:7677
  mapfile: "{cfg_dir}/mapPiezo.yml"  # TO BE UPDATED
  fits_prefix: "OPTI1"
  ctrl_config:
     num_axis:      3
     max_on:        180
""")


""" 
.. automodule::
    :members:  nodealias
"""
from .device import Device, host_mapping, GROUP
from .motor import Motor
from .drot import Drot
from .adc import Adc
from .shutter import Shutter
from .lamp import Lamp
from .piezo import Piezo

from .manager import Manager, open_manager, open_device
from .interface import nodealiasproperty, nodealias, RpcError, Node, RpcNode, NodeAlias, NodeAliasProperty
from .interface import RpcInterface, Interface
from .interface import Int16, Int32, Int64, UInt16, UInt32, UInt64, Float, Double
from .interface import INT  , DINT , LINT , UINT  , UDINT , ULINT  , REAL, LREAL
from .tools import *




from .device import Device, _name_attr, GROUP
from . import trk
from .motor import Motor, axis_type
from . import keys as K
from .manager import Manager # see at the end of file
from .interface import Int16, Int32, nodealiasproperty
from .tools import upload

#                      _              _   
#   ___ ___  _ __  ___| |_ __ _ _ __ | |_ 
#  / __/ _ \| '_ \/ __| __/ _` | '_ \| __|
# | (_| (_) | | | \__ \ || (_| | | | | |_ 
#  \___\___/|_| |_|___/\__\__,_|_| |_|\__|
# 

##### ############
# SUBSTATE
@_name_attr
class SUBSTATE(trk.SUBSTATE):
    pass
### ##############
# errors
@_name_attr
class RPC_ERROR(trk.RPC_ERROR):
    txt = {
        # add more rpc error description here if necessary
    }

### ##############
# RPC error
@_name_attr
class ERROR(trk.ERROR):
    txt = {
        # add more rpc error description here if necessary
    }

### ############# 
# Mode 
@_name_attr
class MODE:
    ENG		= 0
    STAT	= 1
    SKY		= 2
    ELEV	= 3
    USER	= 4
    group = {
        ENG		: GROUP.ENG,
        STAT	: GROUP.STATIC,
        SKY		: GROUP.TRACKING,
        ELEV	: GROUP.TRACKING,
        USER	: GROUP.TRACKING,
    }
    
def mode_parser(mode):
    if isinstance(mode, str):
        if mode not in ['SKY', 'ELEV']:
            raise ValueError('tracking mode must be one of SKY or ELEV got %r'%mode)
        mode = getattr(MODE,mode)
    return Int16(mode)

#  _       _             __                
# (_)_ __ | |_ ___ _ __ / _| __ _  ___ ___ 
# | | '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \
# | | | | | ||  __/ |  |  _| (_| | (_|  __/
# |_|_| |_|\__\___|_|  |_|  \__,_|\___\___|

class DrotStatInterface(Device.StatInterface):
    ERROR = ERROR
    MODE = MODE
    SUBSTATE = SUBSTATE
    
    @nodealiasproperty("track_mode_txt", ["track_mode"])
    def track_mode_txt(self, track_mode):
        return self.MODE.txt.get(track_mode, "UNKNOWN")
    
class DrotCfgInterface(Motor.CfgInterface):
    focus_sign : Int32
    dir_sign   : Int32
    trk_period : Int32
    
# redefine the Method interface to include the proper description of the RpcError
class DrotRpcNode(Device.RpcInterface.RpcNode):
    RPC_ERROR = RPC_ERROR

class DrotRpcInterface(Device.RpcInterface):
    RpcNode = DrotRpcNode
    RPC_ERROR = RPC_ERROR
    ##
    # the type of rpcMethod argument can be defined by annotation
    # All method args types must be defined in a tuple
    rpcMoveAbs    : (float, float)
    rpcMoveRel    : (float, float)
    rpcMoveAngle  : (float,)
    rpcMoveVel    : (float,)
    rpcStartTrack : (mode_parser, float)
    
#      _            _          
#   __| | _____   _(_) ___ ___ 
#  / _` |/ _ \ \ / / |/ __/ _ \
# | (_| |  __/\ V /| | (_|  __/
#  \__,_|\___| \_/ |_|\___\___|
#
class Drot(Motor,trk.Trk):
    SUBSTATE = SUBSTATE
    MODE = MODE 
    ERROR = ERROR
    
    RpcInterface  = DrotRpcInterface
    StatInterface = DrotStatInterface
    CfgInterface  = DrotCfgInterface
    
    def start_track(self, mode, angle=0.0):
        """ Start drot tracking 
        
        Args:
            mode (int, str): tracking mode. Int constant defined in Drot.MODE.SKY, Drot.MODE.ELEV
                             str 'SKY' or 'ELEV' is also accepted
            angle (float): paSky or paPupil depending of the mode
        """
        self.rpc.rpcStartTrack.rcall(mode, angle)
    
    def move_angle(self, angle=0.0):
        """ Move drot to angle in STAT mode """
        self.rpc.rpcMoveAngle.rcall(angle)
    
    def move_abs(self, pos, vel):
        """ Move the drot to an absolute position in ENG mode 
        
        Args:
            pos (float): absolute position
            vel (float):   target velocity for the movement
        """
        self.rpc.rpcMoveAbs.rcall(pos, vel)
    
    def move_rel(self, pos, vel):
        """ Move the drot to a relative position in ENG mode 
        
        Args:
            pos (float): relative position
            vel (float): target velocity for the movement
        """
        self.rpc.rpcMoveRel.rcall(pos, vel)
    
    def move_vel(self, vel):
        """ move drot in velocity 
        
        Args:
           vel (float): target velocity
        """
        self.rpc.rpcMoveVel.rcall( vel)
    
    def stop(self):
        """ Stop drotator motion """
        self.rpc.rpcStop.rcall()
    
Manager.record_new_device_type('Drot', Drot)

Manager.record_cfg_template('Drot', """{name}:
  type: Drot
  interface: Softing
  identifier: PLC1                             # OPCUA Object Identifier
  prefix: MAIN_FAST.drot1                       # TO BE UPDATED OPCUA attribute prefix
  simulated: false
  ignored: false
  address: {address}
  simaddr: opc.tcp://134.171.57.209:4840       # Simulation address
  mapfile: "{cfg_dir}/mapDrot.yml"     # TO BE UPDATED 
  fits_prefix: "DROT1"
  ctrl_config:
    latitude:              -0.429833092
    longitude:             1.228800386
    velocity:              3.0
    min_pos:               -359
    max_pos:               359.0
    axis_type:             CIRCULAR
    active_low_lstop:      false
    active_low_lhw:        false
    active_low_ref:        true
    active_low_index:      false
    active_low_uhw:        true
    active_low_ustop:      false
    brake:                 false
    low_brake:             false
    low_inpos:             false
    backlash:              0.0
    tout_init:             30000
    tout_move:             120000
    tout_switch:           10000
  initialisation:
      sequence: ['FIND_LHW', 'FIND_UHW', 'CALIB_ABS', 'END']
      FIND_LHW:
         value1: 4.0
         value2: 4.0
      FIND_UHW:
         value1: 4.0
         value2: 4.0
      CALIB_ABS:
         value1: 0.0
         value2: 0.0
      END:
         value1: 0.0
         value2: 0.0
  positions:
     posnames: ['ON', 'OFF']
     tolerance: 1                              # Tolerance in UU
     ON: 30
     OFF: 100
""")




from .motor   import MotorCtrl, MotorLine
from .shutter import ShutterCtrl, ShutterLine
from .lamp    import LampCtrl, LampLine
from .adc     import AdcCtrl , AdcLine
from .drot     import DrotCtrl , DrotLine
from .generic import get_widget_constructor



from . import io
from . import keys as K
from .tools import upload 
from .interface import (kjoin, fjoin, Interface, RpcInterface,
                         get_ua_client, connect_ua_client, disconnect_ua_client, 
                         DEFAULT_NAME_SPACE, _name_attr, nodealiasproperty, ksplit, fsplit, fjoin)


# this is a global mapping allowing to change the opc.tcp address on the fly 
# usefull when switching from real PLC to simulated one without having to 
# edit the config files. The key should be the full address to replce (including port)
# e.g. host_mapping = {"opc.tcp://134.171.59.99:4840": "opc.tcp://192.168.1.28:4840"}
host_mapping = {}

_enum = -1 
def _inc(i=None):
    """ number increment to use in frontend 
    
    _inc(0) # reset increment to 0 and return 0 
    _inc()  # increment and return incremented number 
    """
    global _enum
    _enum = _enum+1 if i is None else i
    return _enum

class GROUP:
    """ Constants holder GROUP are used to classify state and substates
    
    This class is not intended to be instancied but only to hold constants
    
    An application if for widget styling for instance : one style per group
    
        SUBSTATE.group[SUBSTATE.NOTOP_NOTREADY] == GROUP.NOK
    """
    # for substate 
    IDL      = "IDL"
    WARNING  = "WARNING"
    ERROR    = "ERROR"
    OK       = "OK"
    NOK      = "NOK"
    BUZY     = "BUZY"
    UNKNOWN  = "UNKNOWN"
    
    # for modes 
    STATIC = "STATIC",
    TRACKING = "TRACKING", 
    ENG = "ENG"
    
    
@_name_attr
class STATE:
    """ constant holder for device STATEs """
    NONE = 0
    NOTOP = 1
    OP = 2
    group = {
       NONE : GROUP.UNKNOWN,
       NOTOP : GROUP.NOK,
       OP    : GROUP.OK,
    }

@_name_attr
class SUBSTATE:
    """ constant holder for device SUBSTATEs """
    # SUBSTATE are specific to each device
    # :TODO: is their common SUBSTATE for each device ? NOTOP_NOTREADY =  100
    #  NOTOP_READY = 101  ?
    NONE = 0
    NOTOP_NOTREADY = 100 # not sure these number are the same accros devices
    NOTOP_READY    = 101
    NOTOP_ERROR    = 199
    
    OP_ERROR =299
    
    UNKNONW = 99999 # add a unknown substate for special cases
    group = {
        NONE                   : GROUP.UNKNOWN,
        NOTOP_NOTREADY         : GROUP.NOK,
        NOTOP_READY            : GROUP.NOK,
        NOTOP_ERROR            : GROUP.ERROR, 
        OP_ERROR               : GROUP.ERROR, 
    }
    
    
    
@_name_attr
class ERROR:
    """ constant holder for device ERRORs """
    OK	                   = 0
    HW_NOT_OP              = 1
    LOCAL                  = 2
    # etc...
    txt = {
    	OK:					  'OK',
    	HW_NOT_OP:			  'ERROR: TwinCAT not OP or CouplerState not mapped.',
    	LOCAL:				  'ERROR: Control not allowed. Motor in Local mode.',
        # etc ...
    }


class DeviceStatInterface(Interface):
    ERROR = ERROR # needed for error_txt alias 
    SUBSTATE = SUBSTATE # needed for substate_txt node alias
    STATE = STATE 
    
    @nodealiasproperty("is_operational", ["state"])
    def is_operational(self, state):
        return state == STATE.OP
    
    @nodealiasproperty("is_ready", ["substate"])
    def is_ready(self, substate):
        return substate == self.SUBSTATE.NOTOP_READY
    
    @nodealiasproperty("is_in_error", ["substate"])
    def is_in_error(self, substate):
        """ -> True is axis is sin error state:  NOP_ERROR or OP_ERROR """
        return substate in [self.SUBSTATE.NOTOP_ERROR, self.SUBSTATE.OP_ERROR]
    
    @nodealiasproperty("substate_txt", ["substate"])
    def substate_txt(self, substate):
        return self.SUBSTATE.txt.get(substate, "not registered substate")
    
    @nodealiasproperty("substate_group", ["substate"])
    def substate_group(self, substate):
        return self.SUBSTATE.group.get(substate, GROUP.UNKNOWN)
    
    
    
    @nodealiasproperty("state_txt", ["state"])
    def state_txt(self, state):
        return self.STATE.txt.get(state, "not registered state")
    
    @nodealiasproperty("state_group", ["state"])
    def state_group(self, state):
        return self.STATE.group.get(state, GROUP.UNKNOWN)
    
    @nodealiasproperty("error_txt", ["error_code"])
    def error_txt(self, error_code):
        return self.ERROR.txt.get(error_code, "No registered Error")
        
class Device:
    """ Base class for Device object 
    
    Most likely a Device will be created by :func:`Device.from_config` from a yml config file
    
    Args:
        key (str): device key (prefix of all nodes)
        config (dict): Dictionary descibing the configuration of the device. (as red from the yml file)
                       See the ESO ICF framework description.
        map (dict): Mapping dictionary of the device
        fits_prefix (str): string prefix for fits keywords                
    """
    STATE = STATE
    SUBSTATE = SUBSTATE
    ERROR = ERROR
    
    _cfg  = None
    _stat = None
    _rpc  = None
    
    CfgInterface  =  Interface
    StatInterface =  DeviceStatInterface
    RpcInterface  =  RpcInterface


    def __init__(self, key, config, map, fits_prefix='', config_file=''):
        self._map = map
        self._config = config
        self._key =  key
        
        
        # what is called prefix in config is actually a name here
        self._fits_key = fjoin(fits_prefix, config.get(K.FITS_PREFIX, ''))
        
        self._config_file = config_file
        # get a client
        # use get_ua_client so we can use an already made connection
        address = config[K.ADDRESS]
        address = host_mapping.get(address, address) # can change address on the fly 
        uaClient = get_ua_client(address)
        self._ua_client = uaClient
        
        
    @classmethod
    def from_config(cl, file_name, cfg_name, key=None, fits_prefix=''):
        """ read a device configuration file and return a :class:`Device` like object to handle it 
        
        Args:
            file_name (str): relative path to a configuration file 
                      The path is relative to one of the directory defined in the
                      $RESPATH environmnet variable
            cfg_name (str): The device name in the config file. 
                    
                For instance in config file like
                    
                    :: 
                         
                        motor1:
                          type: Motor
                          interface: Softing
                          identifier: PLC1
                          etc... 
                
                'motor1' is the `cfg_name`
            
            key (str, optional): The device key. If not given cfg_name is taken
            fits_prefix (str, optional): The fits prefix string for this device
        """
        allconfig = io.load_config(file_name)
        try:
            config = allconfig[cfg_name]
        except KeyError:
            raise ValueError("Device %r does not exists on device definition file %r"%(cfg_name, file_name))
        try:
            mapfile = config[K.MAPFILE]
        except KeyError:
            raise ValueError("Could not find mapfile in device definition %r at %r"%(cfg_name, file_name))
        map_d = io.load_map(mapfile)

        try:
            dtype = config[K.TYPE]
        except KeyError:
            raise ValueError("config file does not define device 'type'")

        try:
            map = map_d[dtype]
        except KeyError:
            raise ValueError("The associated map file does not contain type %r"%dtype)
        
        if key is None:
            key = cfg_name
            
        return cl(key, config, map, fits_prefix=fits_prefix, config_file=file_name)
    
    @property
    def config(self):
        return self._config
        
    @property
    def key(self):
        return self._key
    
    @property
    def prefix(self):
        return ksplit(self._key)[0]

    
    @property
    def name(self):
        return ksplit(self._key)[1]
        
    @property
    def fits_key(self):
        return self._fits_key
    
    @property
    def fits_name(self):
        return fsplit(self._fits_key)[1]

    @property
    def fits_prefix(self):
        return fsplit(self._fits_key)[0]
    
    @property
    def ua_prefix(self):
        return self._config[K.PREFIX]
        
    @property
    def ua_namespace(self):
        return self._config.get(K.NAMESPACE, DEFAULT_NAME_SPACE)

    @property
    def dev_type(self):
        return self._config[K.TYPE]
    
    def connect(self, uaClient=None):
        """ Establish a client connection to OPC-UA server """
        connect_ua_client(self._ua_client, self)
        #self._ua_client.connect()

    def disconnect(self):
        """ disconnect the OPC-UA client """
        disconnect_ua_client(self._ua_client, self)
        # self._ua_client = None
        # return         
        # if self._ua_client is None:
        #     return
        # if self.is_connected():
        #     self._ua_client.disconnect()
            
    def is_connected(self):
        """ Return True if the current device is connected """
        # TODO: check the best way to check a connection from opcua
        if self._ua_client.uaclient and self._ua_client.uaclient._uasocket:
            t = self._ua_client.uaclient._uasocket._thread
            return t and t.is_alive()
        return False
    
    @property
    def map(self):
        """ map dictionary """
        return self._map

    @property
    def cfg(self):
        """ cfg interface containing all the config nodes """
        if self._cfg is None:
            # TODO add the parser functions dictionay
            self._cfg = self.CfgInterface(self.key, self._ua_client, self._map[K.CFG], self.ua_prefix, ua_namespace=self.ua_namespace)
        return self._cfg

    @property
    def stat(self):
        """ stat interface containing all the status Nodes """
        if self._stat is None:
            # TODO add the parser functions dictionay
            self._stat = self.StatInterface(self.key, self._ua_client, self._map[K.STAT], self.ua_prefix,   ua_namespace=self.ua_namespace )
        return self._stat

    @property
    def rpc(self):
        """ rpc interface for all rpcMethods 
        
        Usually the rpc methods are duplicated in the core of the Device class 
        """
        if self._rpc is None:
            # TODO add the parser functions dictionay
            self._rpc = self.RpcInterface(self.key, self._ua_client, self._map[K.RPC], self.ua_prefix, ua_namespace=self.ua_namespace)
        return self._rpc
    
    def clear(self): 
        """ clear some cached values """       
        if self._stat is not None:
            self._stat.clear()
            self._stat = None
        
        if self._cfg is not None:
            self._cfg.clear()
            self._cfg = None
        
        if self._rpc is not None:
            self._rpc.clear()
            self._rpc = None
        
    ## These RPC should be on all devices
    def init(self):
        """ init the device """
        self.rpc.rpcInit.rcall()

    def enable(self):
        """ enable the device """ 
        self.rpc.rpcEnable.rcall()

    def disable(self):
        """ disable the device """
        self.rpc.rpcDisable.rcall()

    def reset(self):
        """ reset the device """
        self.rpc.rpcReset.rcall()

    def get_error_txt(self, errcode):
        """ Get a text description of the given error code number """
        return self.ERROR.txt.get(errcode, 'UNKNOWN ERROR')
    
    def get_rpc_error_txt(self, rpc_errcode):
        """ Get a text description of the given rpc error code number """
        return self.RpcInterface.RPC_ERROR.txt.get(rpc_errcode, 'UNKNOWN RPC ERROR')
    
    def get_configuration(self, **kwargs):
        """ return a node/value pair dictionary ready to be uploaded 
        
        The node/value dictionary represent the device configuration. 
        This is directly use by :func:`Device.configure` method. 
        
        This is a generic configuration dictionary and may not work on all devices. 
        This method need to be updated for each device.   
        
        Args:
            **kwargs : name/value pairs pointing to cfg.name node
                      This allow to change configuration on the fly
                      without changing the config file.             
        
        ::
        
            >>> upload( dev.get_configuration() ) 
        """
        cfg_dict = {self.cfg.get_node(k):v for k,v in self._config.get(K.CTRL_CONFIG, {}).items()}
        cfg_dict.update({self.cfg.get_node(k):v for k,v in kwargs.items()})
        return cfg_dict
    
    def configure(self, **kwargs):
        """ Configure the whole device in the PLC according to what is defined in the config dictionary 
        
        Quick changes on configuration value can be done by keyword where keys must point to a .cfg.name 
        node. Note that the configuration (as written in file) is always loaded first before being 
        overwritten by **kwargs.  
        
        Args:
            **kwargs :  name/value pairs pointing to cfg.name node
                        This allow to quickly change configuration on the fly
                        without changing the config file.
                        Note that when running configure the configuration is always 
                          
        
        this is just:
        
        ::
        
           >>> upload( self.get_condifuration() ) 
        """
        # by default just copy the "ctrl_config" into cfg. This may not work for
        # all devices and should be customized  
        upload(self.get_configuration(**kwargs))
    



from .thread import Thread
from collections import deque, namedtuple, OrderedDict 
from .interface import kjoin, nodealias
from .interface import Node, NodeAlias, NodesReader, NodesWriter
import time
from datetime import datetime, timedelta
import traceback
import sys

__all__ = ['wait', 'Waiter', 'get_utc', 'local_utc', 
           'local_time', 'FifoNodes', 'FifoOneNode', 
            'Downloader', 'Uploader', 'Player', 'Prefixed',
            'suffix_setter', 'setitem', 'download', 'upload']


def get_local_time():
    return time.time()
    
# Add a local_time node alias 
local_time = nodealias("local_time", [])(get_local_time)

def get_utc(delta=0):
    """ return the string utc time ready to be used in some PLC application  
    
        .. note:: 
        
           this is the utctime of the current computer can be shifted from PLC time
           optional delta set a number of seconds to shift from the current computer utc time  
    
    A :class:`NodeAlias` can also be use to get string utc time and local_time to get the computer time 
    
    Args:
        delta: float, optional
              number of seconds to shift the utc from.
    """
    utc = datetime.utcnow()+timedelta(seconds=delta)   
    return datetime.isoformat( utc ).replace("T","-") 

local_utc = nodealias("local_utc", [])(get_utc)

class FifoOneNode(NodeAlias):
    """ This is an :class:`NodeAlias` returning at each get() a :class:`collections.deque` 
    
    This is specially usefull to create graphical scope of on device variable
    
    Args:
       node (:class:`Node`,:class:`NodeAlias`): 
              One single node used to feed the fifo 
       maxlen (int): maximum len of the fifo 
       suffix (string, optional): suffix added to the onput node key to build 
             the alias key default is '_fifo'
    
    Exemple:
        
        ::
        
            > pos_mot = FifoOneNode( mgr.motor1.stat.pos_actual, 1000)
            > pos_mot.get()                                                                                              
            deque([10.504918148932397])
            > pos_mot.get()                                                                                              
            deque([10.504918148932397, 11.704918148932798])
            > pos_mot.get()                                                                                              
            deque([10.504918148932397, 11.704918148932798, 12.75491814893315])
            etc ... 
        
    .. seealso::  
    
        :class:`FifoNodes` : fifo from several nodes
        :class:`NodeAlias`   
    """
    def __init__(self, node, maxlen, suffix="_fifo"):
        key = node.key+suffix
        self.fifo = deque([], maxlen)
        super(FifoOneNode, self).__init__(key, [node])
    
    def fget(self, value):
        self.fifo.append(value)
        return self.fifo

class FifoNodes(NodeAlias):
    """ This is an :class:`NodeAlias` returning at each get a :class:`collections.deque` 
    
    Specially handy for live plot and scope
    
    Args:
       key (string): alias node keyword 
       nodes (list of :class:`Node`,:class:`NodeAlias`): 
              List of nodes to get()  
       maxlen (int): maximum len of the fifo 
    
    Exemple:
        
    In this exxample `local_time` is a builtin node alias returning local time 
    The first motor is moving while the second standstill at 0.0
         
    ::
    
        > from pydevmgr import local_time 
        > nodes = [local_time, mgr.motor1.stat.pos_actual, mgr.motor2.stat.pos_actual]
        > f = FifoNodes( "mot_poses", nodes, 20)
        > f.get()
        deque([(1604931613.136023, 96.44508185106795, 0.0)])
        > f.get()
        deque([(1604931613.136023, 96.44508185106795, 0.0),
        (1604931614.258859, 95.32508185106808, 0.0)])
        > f.get()
        deque([(1604931613.136023, 96.44508185106795, 0.0),
        (1604931614.258859, 95.32508185106808, 0.0),
        (1604931615.451603, 94.13508185106821, 0.0)])
        
        
    """
    def __init__(self, key, nodes, maxlen):        
        super(FifoNodes, self).__init__(key, nodes)
        self.fifo = deque([], maxlen)
                
    def fget(self, *values):
        self.fifo.append(values)
        return self.fifo


def setitem(d,k,v):
    d[k] = v

def download(nodes, data=None):
    """ read node values from servers in one call per server
    
    Args:
        nodes (iterable):
             Iterable of nodes, like [mgr.motor1.stat.pos_actual, mgr.motor2.stat.pos_actual]
        
        data (dict, optional):
             This is mostlikely a dictionary, must define a __setitem__ method
             If given the function return None and update data in place. 
             If data is None the function return a list of values 
             
        
    Returns:
       None, or list : download(nodes) -> return list of values 
                       download(nodes, data) -> return None but update the data dictionary
    
    Exemple:
    
    ::
        
        data = {n.key:n.get() for n in nodes}
    
    Is equivalent, but slower than 
    
    :: 
        
        data = {}
        download(nodes)
        
    The latest is much more efficient because only one call (per server) is done.
    
    data dictionary is optional, if not given values are returned by the function:
    
    ::
     
        pos, error = download([mgr.motor1.stat.pos_actual, mgr.motor1.stat.pos_error])
     
    
    """
    return NodesReader(nodes).read(data)

def upload(node_values):
    """ write node values to the servers
    
    Args:
        node_values (dict):
             Dictionary of node/value pairs like :
                e.g: { motor.cfg.velocity : 4.3 }
    
    .. note:: 
        
        The input dictionary has pairs of node/value and not node.key/value      
    """
    NodesWriter(node_values).write()    



def wait(node_or_func_list, logic="all_true", period=0.1, timeout=60, data=None):
    """ wait until a list of function return True

    Args:
        node_or_func_list (iterable): an iterable of nodes of callable. 
                    wait will wait until all node.get and callable return True (by default)
                    
        logic (str, optional): 
              "all_true"  : stop waiting when all check function return True (default)
              "all_false" : stop waiting when all check function return False
              "any_true"  : stop waiting when one of the check function return True
              "any_false" : stop waiting when one of the check function return False
            
        period (float, optional): default is 0.1 second
            the sleep time in second in between checks
        timeout (float, optional): timeout in second, a RuntimeError is raised if conditions
                         are still false after timeout
        data (None, dict, optional):  If given input nodes values are taken from the 
               data dictionary which is expected to be updated in someother place.

    Exemple:
        Wait until a motor initialised and an other custom function::

            > def camera_ready():
            >   # <- do stuff ->
            >   return True # if camera is ready

            > wait( [motor.is_initialised, camera_ready] )

        Or something like::
        
            > is_arrived = lambda : abs(motor.stat.pos_actual.get()-3.4)<0.01
            > wait( [is_arrived, camera_ready])

    """
    Waiter(node_or_func_list, logic=logic, period=period, timeout=timeout, data=data).wait()

def _all_true(functions, node_get):
    """ all_true(lst, nodes) -> return True if all function and node in list return True """
    for func in functions:
        if not func():
            return False
    return all( node_get() )

def _all_false(functions, node_get):
    """ all_false(lst) -> return True if all function in the list return False """

    for func in functions:
        if func():
            return False
    return all( not v for v in node_get() )
    
def _any_true(functions, node_get):
    """ any_true(lst) -> return True if one of the function in the list return True """
    for func in functions:
        if func():
            return True
    return any( node_get() )

def _any_false(functions, node_get):
    """ any_false(lst) -> return True if one of the function in the list return False """
    for func in functions:
        if not func():
            return True
    return any( not v for v in node_get() )

""" Used in wait to define the logic applied """
_logic_loockup = {
    "all_true"  : _all_true, 
    "all_false" : _all_false, 
    "any_true"  : _any_true, 
    "any_false" : _any_false
}


class Waiter:   
    """ Object use to wait for some condition to be True 
    
    The only method of this object is :method:`Waiter.wait`
    
    Args:
        node_or_func_list (iterable): an iterable of nodes of callable. 
                    waiter.wait will wait until all node.get and callable return True (by default)
        logic (string, optional):
              "all_true"  : stop waiting when all check function return True (default)
              "all_false" : stop waiting when all check function return False
              "any_true"  : stop waiting when one of the check function return True
              "any_false" : stop waiting when one of the check function return False
            
        period : float, optional default is 0.1 second
            the sleep time in second in between checks
        timeout: float, optional second, default is 60
            if time since the call of the function exceed timeout an
            ValueError is raised.
            timeout is in second as well
        data (None, dict, optional):  If given input nodes values are taken from the 
               data dictionary which is expected to be updated in someother place.
        
    Exemple:
        
    ::
        
        >>> wait_moving = Waiter([motor.is_moving])
        >>> wait_moving.wait()
        
    wait return True and can be used as a trigger for a :class:`Download` object :
    
    :: 
            >>> data = {} 
            >>> pos_update = Downloader(data, [motor.stat.pos_actual], triger=wait_moving.wait)
            >>> t = Thread(target=pos_update.runner(0.5))
            >>> t.start()
        
    In the exemple above the thread will update the data dictionary 
    every 0.5 seconds with motor position only when the motor is moving. 
            
    Attributes:
        period (float): conditions are checks every period seconds
        timeout (float): timeout in second, a RuntimeError is raised if conditions
                         are still false after timeout.
    """
    def __init__(self, node_or_func_list, logic="all_true", period=0.1, timeout=60, data=None):        
        nodes = []
        functions = []
        for f in node_or_func_list:
            if isinstance(f, (Node, NodeAlias)):
                nodes.append(f)
            else:
                functions.append(f)
        
        self._functions = functions
            
        
        try:
            check = _logic_loockup[logic]
        except KeyError:
            raise ValueError("undefined logic %r must be one of %s"%(logic, ",".join(_logic_loockup.keys())))
    
        self._check = check
        
        if data is None:
            reader = NodesReader(nodes)
            node_get = reader.read
            
        else:
            def node_get():
                return [data[n.key] for n in nodes]
         
        self._node_get = node_get        
        self.period  = period
        self.timeout = timeout
    
    def wait(self):
        """ run the wait condition """
        
        check = self._check
        node_get = self._node_get
        functions = self._functions 
        timeout = self.timeout
        period = self.period
                
        s_time = time.time()    
        while not check(functions, node_get):
            if (time.time()-s_time)>timeout:
                raise RuntimeError('wait timeout')
            time.sleep(period)
        
        return True
    
def getitem(d,k):
    return d[k]

class Uploader:
    """ An uploader object to upload data to the PLC 
    
    The values to upload is defined in a dictionary of node/value pairs. 
    
    Not sure their is a strong use case for this. Maybe if pydevmgr is used as server instead of client 
    
    Args:
        node_values (dict):
             Dictionary of node/value pairs like :
                e.g: { motor.cfg.velocity : 4.3 }
        callback (callable, optional): callback function after each upload
    
    Example:
        
    ::
    
        >>> values  = {mgr.motor1.velocity: 1.0, mgr.motor2.velocity: 2.0}
        >>> uploader = Uploader(values)
        >>> t = Thread(target=uploader.runner)
        >>> t.start()
        >>> uploader[mgr.motor1.velocity] = 1.4 # will be uploaded at next trhead cycle 
                
    .. seaalso::
    
       :func:`upload`:  equivalent to Uploader(node_values).upload() 
    """
    def __init__(self, node_values=None, callback=None):
        
        node_values = {} if node_values is None else node_values
        
        self.node_values = node_values 
        self.callback = callback
    
    def __has__(self, node):
        return node in self._node_values
        
    def upload(self):
        """ upload the linked node/value dictionaries """
        NodesWriter(self.node_values).write() 
        if self.callback:
            self.callback()
                  
    def run(self, period=1.0, stopsignal=lambda : False, sleepfunc=time.sleep):
        """ Run the upload infinitly or until stopsignal is True 
        
        Args:
            period (float, optional): period of each upload cycle
            stopsignal (callable, optional): A function called at each cycle, if True the loop is break
                       and run returns    
        """
        while not stopsignal():
            s_time = time.time()
            self.upload()
            sleepfunc( max( period-(time.time()-s_time), 0))
    
    def runner(self, period=1.0, stopsignal=lambda : False, sleepfunc=time.sleep): 
        """ return a function to updload 
        
        this is designed in particular to be used in a target Thread
        
        Args:
            period (float, optional): period of each upload cycle
            stopsignal (callable, optional): A function called at each cycle, if True the loop is break
                       and run returns
        
        Example:
            
            ::
                > values  = {mgr.motor1.velocity: 1.0, mgr.motor2.velocity: 2.0}
                > uploader = Uploader(values)
                > t = Thread(target=uploader.runner)
                > t.start()
                > values[mgr.motor1.velocity] = 1.2 # will be updated at next thread cycle
                               
        """           
        def run_func():
            self.run( period=period, sleepfunc=sleepfunc, stopsignal=stopsignal)
        return run_func
    
        
def suffix_setter(suffix, setfunc=setitem):
    def suffixsetfunc(obj, key, value):
        key = key.replace(suffix,'').strip('.') if key.startswith(suffix) else key
        setfunc(obj, key, value)
    return suffixsetfunc


def _dummy_callback(data):
    pass
def _dummy_trigger():
    return True


class Player:
    """ A callable object which can be used in conjonction with Downloader, Uploader 

    > p = Player()
    > d = {}
    > u = Downloader(d, [motor1.stat.pos_actual, motor1.stat.pos_error], trigger=p.wait)
    > t = Thread(target=u.runner(1.0, stopsignal=p.is_stopped))
    > t.start() 
    
    > p.pause() # pause the update of the dictionary
    > p.play()  # restart a paused update
    > p.stop()  # stop the thread

    """
    def __init__(self, period=0.1):
        self._running = True
        self._killed = False
        self._period = period
        
    def __call__(self):
        if self._killed:
            raise StopIteration('')
        return self._running
        
    def play(self):
        self._running = True
        
    def stop(self):
        self._killed = True
        
    def pause(self):
        self._running = False
    
    def is_running(self):
        return self._running
    
    def is_paused(self):
        return not self._killed and not self._running
    
    def wait(self):
        while True:
            if self.is_running():
                return True
            time.sleep(self._period) 
    
    def is_stopped(self):
        return self._killed
    
class Downloader:
    """ object dedicated to download nodes, feed data and run some callback 

    An application can request nodes to be downloaded and callback to be executed after each 
    success download or each failures. 
    
    Args:    
        nodes (iterable of node): An initial, always download, list of nodes 
        data (dict, optional): a dictionary to store the data. If not given, one is created and
                              accessible through the .data attribute
        callback (callable, optional): one single function with signature f(data), if given always 
                                      called after successful download. 
        trigger (callable, optional): a function taking no argument and should return True or False 
                                      If given the "download" if executed only one f() return True. 
                                      Can be used if the download object is running in a thread for instance.
    
    Exemple: 
    
        A dumy exemple, replace the print_pos by a GUI interface for instance:
        
        ::
           
            def print_pos(data):
                "An application"
                print("Position :",  data['pos_actual'], data['pos_error'] )
            
            >>> downloader = Downloader()
            >>> token = download.new_connection()
            >>> downloader.add_node(token, tins.motor1.stat.pos_actual, tins.motor1.stat.pos_error )
            >>> downloader.add_callback(token, print_pos)
            >>> downloader.download() 
            Position : 3.45 0.003
            >>> downloader.data
            {
              'fcs.motor1.pos_actual' : 3.45, 
              'fcs.motor1.pos_error' :  0.003, 
            }
            >>> downloader.disconnect(token) # disconnect the print_pos function and remove 
                                             # the pos_actual, pos_error node from the list of nodes
                                             # to download 
                                     
    """
    def __init__(self,  nodes=None,  data=None, setfunc=setitem, callback=None,
                        trigger=None):
        self._data = {} if data is None else data
        
        nodes = set() if nodes is None else set(nodes)
        callbacks = set()
        failure_callbacks = set() 
        
        
        if callback is None:
            callback = _dummy_callback
        if trigger is None:
            trigger =  _dummy_trigger
        
        self._callback = callback
        self._trigger = trigger
        
        self._dict_nodes = OrderedDict([(None,nodes)])
        self._dict_callbacks = OrderedDict([(None,callbacks)])
        self._dict_failure_callbacks = OrderedDict([(None,failure_callbacks)])
        
        self.setfunc = setfunc
        
        self.trigger = trigger
        self._next_token = 1
        
        self._rebuild_nodes()
        self._rebuild_callbacks()
        self._rebuild_failure_callbacks()
    
    def __has__(self, node):
        return node in self._nodes
    
    @property
    def data(self):
        return self._data
    
    def _rebuild_nodes(self):
        nodes = set()
        for nds in self._dict_nodes.values():
            nodes.update(nds)
        self._nodes = nodes
        self._to_read = NodesReader(nodes)

    def _rebuild_callbacks(self):
        callbacks = set()
        for clbc in self._dict_callbacks.values():
            callbacks.update(clbc)
        self._callbacks = callbacks
    
    def _rebuild_failure_callbacks(self):
        callbacks = set()
        for clbc in self._dict_failure_callbacks.values():
            callbacks.update(clbc)
        self._failure_callbacks = callbacks
    
    def new_connection(self):
        """ add a new app connection 
        
        Return:
           A token, the token and type itself is not relevant, it is just a unique ID to be used in 
                    add_node, add_callback, add_failure_callback, and disconnect methods 
        """
        token = id(self), self._next_token
        self._dict_nodes[token] = set()
        self._dict_callbacks[token] = set()
        self._dict_failure_callbacks[token] = set()
        
        self._next_token += 1
        # self._rebuild_nodes()
        # self._rebuild_callbacks()
        # self._rebuild_failure_callbacks()
        return token
    
    def disconnect(self, token):
        """ Disconnect the App with the given token 
        
        All the nodes used by the app (and not by other connected app) will be removed from the 
        download queue of nodes.
         
        Args:
            token : a Token returned by :func:`Downloader.new_connection`
        """
        if token is None:
            return 
        
        try:
            self._dict_nodes.pop(token)
            self._dict_callbacks.pop(token)
            self._dict_failure_callbacks.pop(token)
        except KeyError:
            pass
        
        self._rebuild_nodes()
        self._rebuild_callbacks()
        self._rebuild_failure_callbacks()
    
    def add_node(self, token, *nodes):
        """ Register nodes to be downloaded for an iddentified app
        
        Args:
            token: a Token returned by :func:`Downloader.new_connection`
            *nodes :  nodes to be added to the download queue, associated to the app
        """   
        self._dict_nodes[token].update(nodes)
        self._rebuild_nodes()
    
    def add_callback(self, token, *callbacks):   
        """ Register callbacks to be executed after each download 
        
        The callback must have the signature f(data)
        
        Args:
            token: a Token returned by :func:`Downloader.new_connection`
            *callbacks :  callbacks to be added to the queue of callbacks, associated to the app
        
        """ 
        self._dict_callbacks[token].update(callbacks)
        self._rebuild_callbacks()
    
    def add_failure_callback(self, token, *callbacks):  
        """ Add one or several callbacks to be executed when a download failed 
        
        The callbacks must have the signature f(e) with e the error grabbed
        
        Args:
            token: a Token returned by :func:`Downloader.new_connection`
            *callbacks :  callbacks to be added to the queue of failure callbacks, associated to the app
        
        """ 
        self._dict_failure_callbacks[token].update(callbacks)
        self._rebuild_failure_callbacks()
    
    def run(self, period=1.0, stopsignal=lambda : False, sleepfunc=time.sleep):
        """ run indefinitely or when stopsignal return True the download 
        
        Args:
            period (float, optional): period between downloads in second
            stopsignal (callable, optional): a function returning True to stop the loop or False to continue
            
        """
        while not stopsignal():
            s_time = time.time()
            self.download()
            sleepfunc( max( period-(time.time()-s_time), 0))
    
    def runner(self, period=1.0, stopsignal=lambda : False, sleepfunc=time.sleep): 
        """ Create a function to run the download in a loop 
        
        It can be the input of a Thread. 
        
        Args:
            period (float, optional): period between downloads in second
            stopsignal (callable, optional): a function returning True to stop the loop or False to continue
        
        Exemple:
            
            >>> downloader = Downloader([mgr.motor1.substate, mgr.motor1.pos_actual])
            >>> t = Thread( target = downloader.runner(period=0.1) )
            >>> t.start()
            
        """       
        def run_func():
            self.run(period=period, sleepfunc=sleepfunc, stopsignal=stopsignal)
        return run_func
    
    def download(self):
        """ Execute a download 
        
        Each nodes on the queue are queried on the OPC-UA server and the .data dictionary is updated
        from new values.
        
        """
        if not self.trigger(): return 
        
        try:
            self._to_read.read(self._data, setfunc=self.setfunc)
        except Exception as e:
            if self._failure_callbacks:
                for func in self._failure_callbacks:
                    func(e)
            else:
                raise e            
        else:
            self._callback(self._data)
            for func in self._callbacks:
                func(self._data)
    
    def get_data(self, prefix=''):
        """ Return a readonly view of the data 
        
        If prefix is given the return object will be limited to items with key
        matching the prefix.  
        
        Args:
           prefix (str, optional): limit the data viewer to a given prefix
        
        Exemple:
            
            ::
                
                > downloader = Downloader([mgr.motor1.substate, mgr.motor2.substate])
                > downloader.download()
                > m1 = downloader.get_data("motor1")
                > m1['pos_actual']
                3.9898
            
        """
        return Prefixed(self._data, prefix)        
    
    def clean_data(self):
        """ Remove to the .data dictionary all keys/value pairs corresponding to nodes not in the queue
        
        """
        node_keys = [n.key for n in self._nodes]
        d = self.data
        for k in list(d):
            if not k in node_keys:
                d.pop(k, None)
            
class Prefixed:
    """ this is a view of a dictionary wit prefixed keys 
    
    d = {"motor1.pos_actual" : 1.5, 
         "motor1.pos_error" : 0.002,
         
         "motor2.pos_actual" : 2.3, 
         "motor2.pos_error" : 0.001}
         
    > dp = Prefixed(d, "motor1") # without the dot
    > dp['pos_actual']
    1.5 
    
    This can be handy when one have to parse for instance a generic "motor" 
    dictionary that should work with any motors.
    
    .. note::
       
       doing dp['pos_actual'] is quite fast has only a prefix is added on the key
            however an iteration on Prefixed dictionary can be long if the root dictionary 
            is large as all keys of the root dictionary have to be checked for prefix.
            It is expected that these Prefixed dictionary will be used mostly as set/get item
    
    .. seealso::  
    
        :func:`download`
        :class:`Downloader`             
    """
    _prefix = ""
    _data = None
    
    def __init__(self, data, prefix):
        self._prefix = prefix
        self._data = data
    
    def __getitem__(self, item):
        return self._data[kjoin(self._prefix, item)]
    
    def __setitem__(self, item, value):
        self._data[kjoin(self._prefix, item)] = value
    
    def __delitem__(self, item):
        del self._data[kjoin(self._prefix, item)]
    
    def __getattr__(self, attr):
        return self._data[kjoin(self._prefix, attr)]
    
    def __has__(self, item):
        return kjoin(self._prefix, item) in self._data 
    
    def update(self, __d__={}, **kwargs):
        for k,v in dict(__d__, **kwargs).iteritems():
            self._data[kjoin(self._prefix, k)] = v
    
    def pop(self, item):
        return self._data.pop(kjoin(self._prefix, item))
    
    def popitem(self, item):
        return self._data.popitem(kjoin(self._prefix, item))
    
    def keys(self):
        """ D.keys() ->iterable on D's root keys with matching prefix
        
        Shall be avoided to use in a :class:`Prefixed` object
        """
        pref = self._prefix+"."
        lp = len(pref)
        for k in self._data.keys():
            if isinstance(k, str) and k.startswith(pref):
                yield k[lp:]
    
    def items(self):
        """D.items() -> iterable on D's root items with matching prefix
        
        Shall be avoided to use in a :class:`Prefixed` object
        """
        pref = self._prefix+"."
        lp = len(pref)
        for k, v in self._data.items():
            if isinstance(k, str) and k.startswith(pref):
                yield k[lp:], v
    
    def values(self):
        """D.values() -> iterable on D's root  values with matching prefix
        
        Shall be avoided to use in a :class:`Prefixed` object
        """
        pref = self._prefix+"."
        for k, v in self._data.items():
            if isinstance(k, str) and k.startswith(pref):
                yield v
    
    def clear(self):
        """ D.clear() -> None.  Remove all items from D root with matching prefix
        
        Shall be avoided to use in a :class:`Prefixed` object
        """
        pref = self._prefix+"."
        for k, v in list(self._data.items()):
            if isinstance(k, str) and k.startswith(pref):
                self._data.pop(k)
    
    
    
